#!/bin/sh

# Defines
INSTALL=0
VERBOSE=0

PRODUCT="IBM JavaPOS 1.7.4 - Component: JavaPOS"
INIT_DIR="/etc/init.d"
RC_INIT_DIR="/etc/rc.d/init.d"
SHARE_DIR="/usr/share/pos"
LIB_DIR="/usr/lib"
UNINSTALL_FILE_LIST="${SHARE_DIR}/.javapos_files"

USB_FS="disable-usbfs-security"
JAVAPOS_SH="ibm-javapos.sh"
PS2_KBD_CONF="ps2kbd.conf"
FLASHDEV="flashdev"


# Install directory list
ETC_DIR="etc"
USR_BIN_DIR="usr/bin"
USR_LIB_DIR="usr/lib"
SBIN_DIR="usr/sbin"

PROFILE_DIR="etc/profile.d"
OPT_DIR="opt"


DIR_LIST="${OPT_DIR} ${USR_LIB_DIR}"


function make_directory()
{
   pdir=""
   dir_parts=$(echo $1 | sed 's/\// /g')

   for path_part in ${dir_parts}
   do
      pdir="${pdir}/${path_part}"
      if [ ! -d ${pdir} ]; then
         mkdir ${pdir}
      fi
   done
}         

function remove_directory()
{
   if [ $# ]; then
      subdirectory=$1
   fi 

   if [ -d "${subdirectory}" ]; then
      for file in ${subdirectory}/*
      do
         if [ -d ${file} ]; then 
            remove_directory ${file}
         else
            if [ ${VERBOSE} -eq 1 ]; then
               echo "Removing /${file}"
            fi

            rm -f /${file}
         fi
      done
   fi
}

function warn_about_previous_install()
{
   echo -e "A previous installation was detected."
   echo -e "We recommend that you uninstall before installing."
   echo -n "Would you like to install anyway?"

   while true
   do
      read continuing
      case "${continuing}" in
         [yY]|yes|YES|Yes)  return 0; break ;;
         [nN]|no|NO|No)     return 1; break ;;
         *)                 echo -n "Please enter \"yes\" or \"no\":"  ;;
      esac
   done 
}

function check_for_previous_install()
{
   if [ -f "${UNINSTALL_FILE_LIST}" ]; then
      if ! warn_about_previous_install ; then
         exit 0
      fi

      rm -f ${UNINSTALL_FILE_LIST}
   fi

   if [ ! -d "${SHARE_DIR}" ]; then
      make_directory ${SHARE_DIR}
   fi
}

function copy_product_files()
{
   for directory in ${DIR_LIST}
   do
      if [ ! -d ${directory} ]; then
         echo -e "Directory ${DIRECTORY} does not exist!\n"
         echo -e "This ${PRODUCT} distribution is incomplete.\n"
         echo -e "Aborting installation..."
         exit 1;
      fi

      if [ ! -d "/${directory}" ]; then
         make_directory "${directory}"
      fi
      
      for files in ${directory}
      do
         if [ ${VERBOSE} -eq 1 ]; then
            echo -e "Installing the files in ${files}"
         fi

         cp -pR ${files}/* /${directory} > /dev/null 2>&1
      done
   done
}

function add_to_etc_dir()
{
   # Install the init script
   init_dir=""
   if [ -d ${INIT_DIR} ]; then
      init_dir=${INIT_DIR}
   elif [ -d ${RC_INIT_DIR} ]; then
      init_dir=${RC_INIT_DIR}
   fi

   #copy files
   cp -f etc/init.d/${USB_FS} ${init_dir}
   cp -f etc/init.d/${FLASHDEV} ${init_dir}
   
   cp -f ${ETC_DIR}/${PS2_KBD_CONF} /${ETC_DIR}
   cp -f ${PROFILE_DIR}/${JAVAPOS_SH} /${PROFILE_DIR}

   # Add init script links
  
   if [ ${VERBOSE} -eq 1 ]; then
      echo "Creating runlevel symbolic links for init script"
   fi

   /sbin/chkconfig --add ${FLASHDEV}
 
   if [ $? != 0 ]; then
      echo "Could not add init script"
      exit 1;
   fi
}

function remove_from_etc_dir()
{
   # Install the init script
   init_dir=""
   if [ -d ${INIT_DIR} ]; then
      init_dir=${INIT_DIR}
   elif [ -d ${RC_INIT_DIR} ]; then
      init_dir=${RC_INIT_DIR}
   fi

   #remove files
   if [ -f ${init_dir}/${USB_FS} ]; then
      if [ ${VERBOSE} -eq 1 ]; then
          echo -e "Installing the files in ${init_dir}/${USB_FS} "
      fi
   
      rm -f ${init_dir}/${USB_FS}
   fi
   
   if [ -f ${init_dir}/${FLASHDEV} ]; then
      if [ ${VERBOSE} -eq 1 ]; then
          echo -e "Installing the files in ${init_dir}/${FLASHDEV}"
      fi
   
      rm -f ${init_dir}/${FLASHDEV}
   fi
   
   if [ -f /${ETC_DIR}/${PS2_KBD_CONF} ]; then
      if [ ${VERBOSE} -eq 1 ]; then
          echo -e "Installing the files in /${ETC_DIR}/${PS2_KBD_CONF}"
      fi
   
      rm -f /${ETC_DIR}/${PS2_KBD_CONF}
   fi
   
   if [ -f /${PROFILE_DIR}/${JAVAPOS_SH} ]; then
      if [ ${VERBOSE} -eq 1 ]; then
          echo -e "Installing the files in /${PROFILE_DIR}/${JAVAPOS_SH}"
      fi
      
      rm -f /${PROFILE_DIR}/${JAVAPOS_SH}
   fi

}



function remove_lib_files()
{
   for file in ${LIB_DIR}/libaip*
   do
      if [ -h ${file} ]; then
         if [ ${VERBOSE} -eq 1 ]; then
            echo "Removing ${file}"
         fi
         rm -f ${file}
      fi
   done
   /sbin/ldconfig
}


function create_symlinks()
{
   if [ ${VERBOSE} -eq 1 ]; then
      echo -e "\nEstablishing symbolic links for ${PRODUCT} shared libraries\n"
      /sbin/ldconfig -v 2>&1 | grep "libaip"
   else
      /sbin/ldconfig
   fi
}



function remove_pos_directories()
{
   for directory in ${DIR_LIST}
   do
      pos_dir=$(expr ${directory} : '\(.*pos\).*')

      if [ "${pos_dir}" != "" ]; then
         if [ -d /${pos_dir} ]; then
            if [ ${VERBOSE} -eq 1 ]; then
               echo "Removing /${pos_dir}"
            fi

            rm -Rf /${pos_dir}
         fi
      fi
   done
}

function remove_installed_files()
{
   curr_dir=$(pwd)

   # Try to use our file "database".
   if [ -a ${UNINSTALL_FILE_LIST} ]; then
      while read line
      do
         if [ ${VERBOSE} -eq 1 ]; then
            echo "Removing ${line}"
         fi

         rm -f ${line}
      done < ${UNINSTALL_FILE_LIST}
   elif expr "${curr_dir}" : ".*ibmposs-linux-[0-9]*\.[0-9]*\.[0-9]*.*" >/dev/null
   then
      for directory in `ls`
      do
         if [ ! -f ${directory} ]; then
            remove_directory ${directory}
         fi
      done
   else   
      echo "Could not determine which files to remove!"
   fi
}


function create_file_list()
{
   subdirectory=$1
 
   if [ -d ${subdirectory} ]; then
      for file in ${subdirectory}/*
      do
         if [ -d ${file} ]; then
            create_file_list ${file}
         else
            echo "/${file}" >> ${UNINSTALL_FILE_LIST}
            if [ ${VERBOSE} -eq 1 ]; then
               echo -e "/${file}"
            fi
         fi
      done
   fi              
}

function prepare_uninstall()
{
   if [ ! -d "${SHARE_POS}" ]; then
      make_directory "${SHARE_POS}"
   fi

   if [ ${VERBOSE} -eq 1 ]; then
      echo -e "Creating an uninstallation file list in ${UNINSTALL_FILE_LIST}\n"
      echo -e "Package files:\n"
   fi

   for directory in ${DIR_LIST}
   do   
      if [ ! -f ${directory} ]; then
         create_file_list ${directory}
      fi       
   done
}

 
function install_product()
{
   # Do a perfunctory check and warn for previous installations.
   check_for_previous_install

   prepare_uninstall
   copy_product_files
   create_symlinks
   add_to_etc_dir
}

function uninstall_product()
{

   # Remove files installed from package
   remove_installed_files

   # Let's remove ldconfig symbolic links now
   remove_lib_files
   #remove_pos_directories
   remove_from_etc_dir 
}   

function help_opt()
{
   echo -e "The proper usage is as follows :
           ./install_pos.sh [-i,--install -u,--uninstall, -h,--help]\n\
         	-i,--install    Installs the POS subsystem
                -u,--uninstall  Uninstalls the POS subsystem
                -v,--verbose    Print messages about the install or uninstall
                -h,--help       print this help message
         "
   exit 1
}
 
function parse_opts()
{
   # The option parsing is portable to systems without getopt installed.
   if which getopt > /dev/null; then
      INSTALL_OPTS=`getopt -o iuvh -l install,uninstall,verbose,help  -- "$@"` 

      eval set -- "$INSTALL_OPTS"
   fi

   while [ $# -gt 0 ]
   do
      case "$1" in
         -i|--install)   if [ ${INSTALL} -ne 0 ]; then 
                            echo "You cannot install and uninstall!"
                            help_opt
                         fi
                         INSTALL=1
                         shift;;
         -u|--uninstall) if [ ${INSTALL} -ne 0 ]; then 
                            echo "You cannot install and uninstall!"
                            help_opt
                         fi
                         INSTALL=2
                         shift;; 
         -v|--verbose)   VERBOSE=1
                         shift;;
         -h|--help)      help_opt
                         shift 2;;
         --)             shift;
                         break ;;
         *)              help_opt
                         exit 1;;
      esac
   done                 
}

parse_opts $@

# Should the product be installed, uninstalled, or just show the help
if [ "${INSTALL}" -eq 1 ]; then

   if [ ${VERBOSE} -eq 1 ]; then
      echo -e "Installing the ${PRODUCT}\n"
   fi

   install_product

elif [ "${INSTALL}" -eq 2 ]; then

   if [ ${VERBOSE} -eq 1 ]; then
      echo -e "Uninstalling the ${PRODUCT}\n\n"
   fi

   uninstall_product

else

   help_opt 

fi
