#!/bin/bsh

# Defines
INSTALL=0
VERBOSE=0

PRODUCT="IBM Point of Sale Subsystem for Linux"
KERNEL_VERSION=""
KERNEL_RELEASE=""
KERNEL_MAJOR=""
INIT_DIR="/etc/init.d"
RC_INIT_DIR="/etc/rc.d/init.d"
SHARE_DIR="/usr/share/pos"
LIB_DIR="/usr/lib"
UNINSTALL_FILE_LIST="${SHARE_DIR}/.pos_files"
AIPSTART="aipstart"

# Install directory list
USR_BIN_DIR="usr/bin"
DOC_DIR="usr/doc/pos"
INCLUDE_DIR="usr/include/pos"
USR_LIB_DIR="usr/lib"
SBIN_DIR="usr/sbin"
SHARE_POS_DIR="usr/share/pos"
VAR_LIB_DIR="var/lib/misc"

DIR_LIST="${USR_BIN_DIR} ${DOC_DIR} ${INCLUDE_DIR} ${USR_LIB_DIR} ${SBIN_DIR} ${SHARE_POS_DIR} ${VAR_LIB_DIR}"

ETC_DIR="etc"
DEV_DIR="/dev"
UDEV_DEV_DIR="/etc/udev/devices"
AIPSYS_RES="aipsys.res"
AIPSYS_CONF="aipsys.conf"
AIPTSTR_RES="aiptstr.res"

# Character devices
AIPDCS="aipdcs:245:0:All"
AIPIKBPS="aipikbps:246:0:All"
AIPBCD="aipbcd:247:0:All"
AIPTCHMOUSE="aiptchmouse:248:0:2.4"
AIPMTN="aipmtn:249:0:All"
TOUCH1="input/touch0:13:248:2.4"
TOUCH2="input/touch1:13:249:2.4"
TOUCH3="input/touch:13:250:2.4"
DEV_LIST="${AIPDCS} ${AIPIKBPS} ${AIPBCD} ${AIPTCHMOUSE} ${AIPMTN} ${TOUCH1} ${TOUCH2} ${TOUCH3}"

# Log files
VAR_LOG="/var/log"
AIPANPOS_LOG=aipanpos.log
AIPCTRL_LOG_0=aipctrl.log
AIPCTRL_LOG_1=aipctrl.log.1
AIPCTRL_LOG_2=aipctrl.log.2
AIPCTRL_LOG_3=aipctrl.log.3
AIPDEBUG_LOG=aipdebug.log
AIPTRACE_LOG=aiptrace.log
LOG_LIST="${AIPANPOS_LOG} ${AIPCTRL_LOG_0} ${AIPCTRL_LOG_1} ${AIPCTRL_LOG_2} ${AIPCTRL_LOG_3} ${AIPDEBUG_LOG} ${AIPTRACE_LOG}"


function get_kernel_info()
{
   KERNEL_INFO=$(uname -r)
   KERNEL_VERSION=$(echo ${KERNEL_INFO} | cut -f1 -d.)
   KERNEL_RELEASE=$(echo ${KERNEL_INFO} | cut -f2 -d.)
   KERNEL_MAJOR=$(echo ${KERNEL_INFO} | cut -f3 -d. | sed "s/\-.*//")
}

function make_directory()
{
   pdir=""
   dir_parts=$(echo $1 | sed 's/\// /g')

   for path_part in ${dir_parts}
   do
      pdir="${pdir}/${path_part}"
      if [ ! -d ${pdir} ]; then
         mkdir ${pdir}
      fi
   done
}         

function remove_directory()
{
   if [ $# ]; then
      subdirectory=$1
   fi 

   if [ -d "${subdirectory}" ]; then
      for file in ${subdirectory}/*
      do
         if [ -d ${file} ]; then 
            remove_directory ${file}
         else
            if [ ${VERBOSE} -eq 1 ]; then
               echo "Removing /${file}"
            fi

            rm -f /${file}
         fi
      done
   fi
}

function warn_about_previous_install()
{
   echo -e "A previous installation was detected."
   echo -e "We recommend that you uninstall before installing."
   echo -n "Would you like to install anyway?"

   while true
   do
      read continuing
      case "${continuing}" in
         [yY]|yes|YES|Yes)  return 0; break ;;
         [nN]|no|NO|No)     return 1; break ;;
         *)                 echo -n "Please enter \"yes\" or \"no\":"  ;;
      esac
   done 
}

function check_for_previous_install()
{
   if [ -f "${UNINSTALL_FILE_LIST}" ]; then
      if ! warn_about_previous_install ; then
         exit 0
      fi

      rm -f ${UNINSTALL_FILE_LIST}
   fi

   if [ ! -d "${SHARE_DIR}" ]; then
      make_directory ${SHARE_DIR}
   fi
}

function copy_product_files()
{
   for directory in ${DIR_LIST}
   do
      if [ ! -d ${directory} ]; then
         echo -e "Directory ${DIRECTORY} does not exist!\n"
         echo -e "This ${PRODUCT} distribution is incomplete.\n"
         echo -e "Aborting installation..."
         exit 1;
      fi

      if [ ! -d "/${directory}" ]; then
         make_directory "${directory}"
      fi
      
      for files in ${directory}
      do
         if [ ${VERBOSE} -eq 1 ]; then
            echo -e "Installing the files in ${files}"
         fi

         cp -pR ${files}/* /${directory} > /dev/null 2>&1
      done
   done
}

function update_etc_directory()
{
   # Install the init script
   init_dir=""
   if [ -d ${INIT_DIR} ]; then
      init_dir=${INIT_DIR}
   elif [ -d ${RC_INIT_DIR} ]; then
      init_dir=${RC_INIT_DIR}
   fi

   cp -f etc/init.d/${AIPSTART} ${init_dir}

   if [ ! -f /${ETC_DIR}/${AIPSYS_CONF} ]; then
      cp  -f ${ETC_DIR}/${AIPSYS_CONF} /${ETC_DIR}
   fi

   if [ ! -f /${ETC_DIR}/${AIPSYS_RES} ]; then
      cp  -f ${ETC_DIR}/${AIPSYS_RES} /${ETC_DIR}
   fi

   if [ ! -f /${ETC_DIR}/${AIPTSTR_RES} ]; then
      cp  -f ${ETC_DIR}/${AIPTSTR_RES} /${ETC_DIR}
   fi
}


  
function add_to_system_boot()
{
   # Install the init script
   init_dir=""
   if [ -d ${INIT_DIR} ]; then
      init_dir=${INIT_DIR}
   elif [ -d ${RC_INIT_DIR} ]; then
      init_dir=${RC_INIT_DIR}
   fi

   if [ ${VERBOSE} -eq 1 ]; then
      echo "Creating runlevel symbolic links for init script"
   fi

	if [ -f /sbin/insserv ]; then
		/sbin/insserv ${init_dir}/${AIPSTART}
	elif [ -f /sbin/chkconfig ]; then
		/sbin/chkconfig --add ${AIPSTART}
	fi
 
   if [ $? != 0 ]; then
      echo "Could not add init script"
      exit 1;
   fi
}

function create_device_nodes()
{
   os="${KERNEL_VERSION}_${KERNEL_RELEASE}"

   for file in ${DEV_LIST}
   do
      devnode=$(echo ${file} | cut -f1 -d:)
      major=$(echo ${file} | cut -f2 -d:)
      minor=$(echo ${file} | cut -f3 -d:)
      version=$(echo ${file} | cut -f4 -d: | sed "s/\./_/g")

      if [ "${version}" = "All" ] || [ "${version}" = "2.4" ]; then
         if [ ! -c "${DEV_DIR}/${devnode}" ]; then
            if [ ${VERBOSE} -eq 1 ]; then
               echo "Creating device node ${DEV_DIR}/${devnode}"
            fi

            /bin/mknod ${DEV_DIR}/${devnode} c ${major} ${minor}
         fi   

         #  Support udev static device nodes
         if [ -d "${UDEV_DEV_DIR}" ]; then
            if [ ! -c "${UDEV_DEV_DIR}/${devnode}" ]; then
               if [ ${VERBOSE} -eq 1 ]; then
                  echo "Creating device node ${UDEV_DEV_DIR}/${devnode}"
               fi

               name=$(basename ${UDEV_DEV_DIR}/${devnode})

               if [ ! -d ${name} ]; then
                  mkdir ${name} 2>/dev/null
               fi

               /bin/mknod ${UDEV_DEV_DIR}/${devnode} c ${major} ${minor}
            fi   
         fi   
      fi   
   done
}

function remove_device_nodes()
{
   os="${KERNEL_VERSION}_${KERNEL_RELEASE}"

   for file in ${DEV_LIST}
   do
      devnode=$(echo ${file} | cut -f1 -d:)
      version=$(echo ${file} | cut -f4 -d: | sed "s/\./_/g")

      if [ "${version}" = "All" ] || [ "${version}" = "2_4" ]; then
         if [ -a ${DEV_DIR}/${devnode} ]; then
            if [ ${VERBOSE} -eq 1 ]; then
               echo "Removing device node ${DEV_DIR}/${devnode}"
            fi

            rm -f ${DEV_DIR}/${devnode}
         fi   

         #  Support udev static device nodes
         if [ -d "${UDEV_DEV_DIR}" ]; then
            if [ ! -c "${UDEV_DEV_DIR}/${devnode}" ]; then
               if [ ${VERBOSE} -eq 1 ]; then
                  echo "Removing device node ${UDEV_DEV_DIR}/${devnode}"
               fi

               rm -f ${UDEV_DEV_DIR}/${devnode}
            fi   
         fi   
      fi   
   done
}

function remove_lib_files()
{
   for file in ${LIB_DIR}/libaip*
   do
      if [ -h ${file} ]; then
         if [ ${VERBOSE} -eq 1 ]; then
            echo "Removing ${file}"
         fi
         rm -f ${file}
      fi
   done
   /sbin/ldconfig
}

function remove_log_files()
{
   # Clean up /var/log
   for file in ${LOG_LIST}
   do
      if [ -f ${VAR_LOG}/${file} ]; then
         if [ ${VERBOSE} -eq 1 ]; then
            echo "Removing ${VAR_LOG}/${file}" 
         fi

         if [ -f ${VAR_LOG}/${file} ]; then
            rm -f ${VAR_LOG}/${file}
         fi
      fi          
   done
}

function stop_active_files()
{
   # Shut down the IBM POS Subsystem
   init_dir=""
   if [ -d ${INIT_DIR} ]; then
      init_dir=${INIT_DIR}
   elif [ -d ${RC_INIT_DIR} ]; then
      init_dir=${RC_INIT_DIR}
   fi

   if [ -a ${init_dir}/${AIPSTART} ]; then
      ${init_dir}/${AIPSTART} stop
   fi
   
   file=$(echo ${AIPIKBPS} | cut -f1 -d:)

   # Remove the keyboard driver
   if [ "$(grep ${file} /proc/modules)" ]; then   
       if [ ${VERBOSE} -eq 1 ]; then
          echo "Removing ${PRODUCT} System keyboard driver" 
       fi

       for I in 0 1 2 3 4 5 6 7 8 9
       do
           rmmod ${file} >/dev/null 2>&1
           if [ $(grep ${file} /proc/modules) ]; then
               usleep 200000
	   else
	       break
	   fi
       done
   fi
}

function create_symlinks()
{
   if [ ${VERBOSE} -eq 1 ]; then
      echo -e "\nEstablishing symbolic links for ${PRODUCT} shared libraries\n"
      /sbin/ldconfig -v 2>&1 | grep "libaip"
   else
      /sbin/ldconfig 2>&1 | grep "libaip"
   fi
}

function create_developer_links()
{
   # Add symlink for developers
   if [ ! -h ${LIB_DIR}/libaiplib.so ]; then
      ln -s $(ls ${LIB_DIR}/libaiplib.*.*.*) ${LIB_DIR}/libaiplib.so
   fi
}

function remove_from_system_boot()
{
   if [ ${VERBOSE} -eq 1 ]; then
      echo "Removing runlevel symbolic links for init script"
   fi

	init_dir=""
	if [ -d ${INIT_DIR} ]; then
		init_dir=${INIT_DIR}
	elif [ -d ${RC_INIT_DIR} ]; then
		init_dir=${RC_INIT_DIR}
	fi

	if [ -a ${init_dir}/${AIPSTART} ]; then
		if [ -f /sbin/insserv ]; then
			/sbin/insserv -r ${init_dir}/${AIPSTART}
		elif [ -f /sbin/chkconfig ]; then
			/sbin/chkconfig --del ${AIPSTART}
		fi
	fi
 
   if [ $? != 0 ]; then
      echo "Could not remove init script"
      exit 1;
   fi
}

function remove_pos_directories()
{
   for directory in ${DIR_LIST}
   do
      pos_dir=$(expr ${directory} : '\(.*pos\).*')

      if [ "${pos_dir}" != "" ]; then
         if [ -d /${pos_dir} ]; then
            if [ ${VERBOSE} -eq 1 ]; then
               echo "Removing /${pos_dir}"
            fi

            rm -Rf /${pos_dir}
         fi
      fi
   done
}

function remove_installed_files()
{
   curr_dir=$(pwd)

   # Try to use our file "database".
   if [ -a ${UNINSTALL_FILE_LIST} ]; then
      while read line
      do
         if [ ${VERBOSE} -eq 1 ]; then
            echo "Removing ${line}"
         fi

         rm -f ${line}
      done < ${UNINSTALL_FILE_LIST}
   elif expr "${curr_dir}" : ".*ibmposs-linux-[0-9]*\.[0-9]*\.[0-9]*.*" >/dev/null
   then
      for directory in $(ls)
      do
         if [ ! -f ${directory} ]; then
            remove_directory ${directory}
         fi
      done
   else   
      echo "Could not determine which files to remove!"
   fi
}


function create_file_list()
{
   subdirectory=$1
 
   if [ -d ${subdirectory} ]; then
      for file in ${subdirectory}/*
      do
         if [ -d ${file} ]; then
            create_file_list ${file}
         else
            echo "/${file}" >> ${UNINSTALL_FILE_LIST}
            if [ ${VERBOSE} -eq 1 ]; then
               echo -e "/${file}"
            fi
         fi
      done
   fi              
}

function prepare_uninstall()
{
   if [ ! -d "${SHARE_POS}" ]; then
      make_directory "${SHARE_POS}"
   fi

   if [ ${VERBOSE} -eq 1 ]; then
      echo -e "Creating an uninstallation file list in ${UNINSTALL_FILE_LIST}\n"
      echo -e "Package files:\n"
   fi

   for directory in ${DIR_LIST}
   do   
      if [ ! -f ${directory} ]; then
         create_file_list ${directory}
      fi       
   done
}

 
function install_product()
{
   # Do a perfunctory check and warn for previous installations.
   check_for_previous_install

   prepare_uninstall
   copy_product_files
   create_symlinks
	create_developer_links
   # create_device_nodes
   update_etc_directory
   add_to_system_boot
}

function uninstall_product()
{
   stop_active_files

	remove_from_system_boot
   remove_installed_files

   # Let's remove ldconfig symbolic links now
   remove_lib_files
   remove_device_nodes
   remove_log_files
   remove_pos_directories
}

function help_opt()
{
   echo -e "The proper usage is as follows :
           ./install_pos.sh [-i,--install -u,--uninstall, -h,--help]\n\
         	-i,--install    Installs the POS subsystem
                -u,--uninstall  Uninstalls the POS subsystem
                -v,--verbose    Print messages about the install or uninstall
                -h,--help       print this help message
         "
   exit 1
}
 
function parse_opts()
{
   # The option parsing is portable to systems without getopt installed.
   if which getopt > /dev/null; then
      INSTALL_OPTS=$(getopt -o iuvh -l install,uninstall,verbose,help  -- "$@") 

      eval set -- "$INSTALL_OPTS"
   fi

   while [ $# -gt 0 ]
   do
      case "$1" in
         -i|--install)   if [ ${INSTALL} -ne 0 ]; then 
                            echo "You cannot install and uninstall!"
                            help_opt
                         fi
                         INSTALL=1
                         shift;;
         -u|--uninstall) if [ ${INSTALL} -ne 0 ]; then 
                            echo "You cannot install and uninstall!"
                            help_opt
                         fi
                         INSTALL=2
                         shift;; 
         -v|--verbose)   VERBOSE=1
                         shift;;
         -h|--help)      help_opt
                         shift 2;;
         --)             shift;
                         break ;;
         *)              help_opt
                         exit 1;;
      esac
   done                 
}

parse_opts $@

get_kernel_info

# Should the product be installed, uninstalled, or just show the help
if [ "${INSTALL}" -eq 1 ]; then

   if [ ${VERBOSE} -eq 1 ]; then
      echo -e "Installing the ${PRODUCT}\n"
   fi

   install_product

elif [ "${INSTALL}" -eq 2 ]; then

   if [ ${VERBOSE} -eq 1 ]; then
      echo -e "Uninstalling the ${PRODUCT}\n\n"
   fi

   uninstall_product

else

   help_opt 

fi
