/* 
 * Module Name:     aipbcd-2.6.c
 *
 * Description:     Cash Drawer Driver for Linux 2.6.X
 *
 *  (C) Copyright 2004,2005 International Business Machines Corporation.
 *  All right reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1.  Redistributions of source code must retain the above copyright notice,
 *     this list of conditions and the following disclaimer.
 *
 * 2.  Redistributions in binary form must reproduce the above copyright notice,
 *     this list of conditions and the following disclaimer in the documentation
 *     and/or other materials provided with the distribution.
 *
 * 3.  The name of the author may not be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Should you need to contact me, the author, you can do so either by
 * e-mail - mail your message to <nobody@us.ibm.com>.
 */

/* System Headers */
#include <linux/version.h>
#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h> 
#include <linux/compiler.h>
#include <linux/pci.h>
#include <linux/init.h>
#include <linux/ioport.h>
#include <linux/delay.h>
#if LINUX_VERSION_CODE > KERNEL_VERSION(2, 5, 0)
#include <linux/cdev.h>
#include <linux/moduleparam.h>
#endif
#include <linux/timer.h>
#include <linux/interrupt.h>
#include <asm/uaccess.h>

/* Local configuration constants */
#ifndef __LINUX__
#define __LINUX__
#endif
#ifndef __UNIX__
#define __UNIX__
#endif

/* Local Headers */
#include "aipkernel.h"

MODULE_AUTHOR("IBM Retail Store Solutions");
MODULE_DESCRIPTION("IBM Cash Drawer Driver");
MODULE_LICENSE("GPL compatible");

typedef struct {
	unsigned long SlotDrawers;
	unsigned long CDRegAddr;
	int boardSize;			/* I/O space length  */
	pid_t cd_pid;
	struct pci_dev *pcidev;
	unsigned short OpenCount;
} BOARDDATA;



/* Local Constants */
#define DEVDRVR_ID "aipbcd"

#define DEFAULT_MAJOR_DEV 247


#define IBM_VENDOR_ID      	0x1014
#define NM_VENDOR_ID		0x9710
#define NM_DEVICE_ID		0x9835
#define SP500_DEVICE_ID		0x299
#define SP500_MODELID_1		0x34
#define SP500_MODELID_2		0x42
#define SP500_31X_MODELID_3	0x59
#define SP500_32X_MODELID_3	0x57

#define CD2_CONNECTED		0x10
#define CD2_OPEN		0x20
#define CD1_CONNECTED		0x40
#define CD1_OPEN		0x80

#define OPEN_CMD_SIZE 10
unsigned char OpenCDCmd[OPEN_CMD_SIZE] =
		{ 0xFF, 0x7F, 0xD5, 0x55, 0xAA, 0x2A, 0xFE, 0x7E, 0xFF, 0x7F};

/* Function Prototypes */
static int __init aipbcd_init(void);
static void __exit aipbcd_exit (void);
static int aipbcd_open(struct inode *inode, struct file *file);
static ssize_t aipbcd_read(struct file *file, char *buf, size_t count,
			   loff_t *ppos);
static ssize_t aipbcd_write(struct file *file, const char *buf, size_t count,
			    loff_t *ppos);
static int aipbcd_release(struct inode *inode, struct file *file);
static int aipbcd_ioctl(struct inode *inode, struct file *file,
			unsigned int cmd, unsigned long arg);


/* PCI Cash Drawer-specific functions */
static int disable_async_comm(char *pBuffer);
static int enable_async_comm(char *pBuffer);
static int get_slot_info(char *pBuffer);
static int get_cd_status(char *pBuffer);
static int open_cd_drawer(char *pBuffer);
static int set_cd_interface(char *pBuffer);


module_init(aipbcd_init);
module_exit(aipbcd_exit);


/* static global variables */
static int aipbcd_major = DEFAULT_MAJOR_DEV;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 0)
MODULE_PARM(aipbcd_major, "i");
#else
module_param_named(major, aipbcd_major, int, 0644);
#endif


static BOARDDATA BoardData;

#ifdef DEBUG
#define DPRINTK(fmt, args...) printk(KERN_DEBUG "%s(%s): " fmt, DEVDRVR_ID, __FUNCTION__ , ## args)
#else
#define DPRINTK(fmt, args...)
#endif


/* The various file operations we support. */
static struct file_operations aipbcd_fops = {
	.owner          = THIS_MODULE,
	.read           = aipbcd_read,
	.write          = aipbcd_write,
	.ioctl          = aipbcd_ioctl,
	.open           = aipbcd_open,
	.release        = aipbcd_release,
};


/* aipbcd_init: Init device driver 
 * ************************************************************************
 *
 *  Function Name:  aipbcd_init
 *
 *  Purpose:        Initialize this device driver (called at insmod/modprobe
 *                  time)
 *
 *  Description:    Find PCI and/or ISA (4694) adapters
 *                  and get configuration information (2K buffer address,  
 *                  NVRAM address, IRQ). Register interrupt handler and 
 *                  send initial POR.
 *
 *  Input:          none
 *
 *  Output:         0  -- success
 *                  -1 -- error
 *
 * ************************************************************************
 */

static int __init aipbcd_init(void)
{
	int err = 0;
	int rc = 0;
	unsigned char *pmem;
	unsigned char id_byte1;
	unsigned char id_byte2;
	unsigned char id_byte3;
	unsigned char cdConfig = 0x00;
	struct pci_dev *pcidev = 0;

	/* Initialize BoardData */
	memset(&BoardData, 0, sizeof(BOARDDATA));
	BoardData.boardSize = 2;

	if ((pmem = (unsigned char *)ioremap(0xFE000, 3)) == 0) {
		DPRINTK("Unable to map memory at 0xFE000\n");
		rc = -ENODEV;
		goto out;
	}

	id_byte1 = readb(pmem);
	id_byte2 = readb(pmem + 1);
	id_byte3 = readb(pmem + 2);
	iounmap(pmem);

	DPRINTK("ID Bytes= %#x %#x %#x\n", id_byte1, id_byte2, id_byte3);

	if (!((id_byte1 == SP500_MODELID_1) &&
	      (id_byte2 == SP500_MODELID_2) &&
	      (id_byte3 == SP500_32X_MODELID_3))) {

	        rc = -ENODEV;
		goto out;
	}

	DPRINTK("SP500 Model 32x\n");

	/* Register as a device with kernel. */
	if ((rc = register_chrdev(aipbcd_major, DEVDRVR_ID, &aipbcd_fops))) {
		printk(KERN_ERR "%s: Failure to register device, error=%d\n",
		       DEVDRVR_ID, -err);
                rc = -ENODEV;
		goto err_out;
	}

	DPRINTK("Device registered with kernel\n");

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 9)
	pcidev = pci_find_subsys(NM_VENDOR_ID, NM_DEVICE_ID, 
				 IBM_VENDOR_ID, SP500_DEVICE_ID, pcidev);
#else
	pcidev = pci_get_subsys(NM_VENDOR_ID, NM_DEVICE_ID, 
				IBM_VENDOR_ID, SP500_DEVICE_ID, pcidev);
#endif

	if (pcidev == NULL) {
		DPRINTK("Device not found\n");
		rc = -ENODEV;
		goto err_out;
	}

	BoardData.pcidev = pcidev;
	DPRINTK("SP500 with Cash Drawer found\n");

	if (pci_enable_device(pcidev) != 0) {
		DPRINTK("Unable to enable device\n");
		rc = -ENODEV;
		goto err_out;
	}

	/* Get CD Register address */
	pci_read_config_dword(pcidev, PCI_BASE_ADDRESS_2,
			      (u32 *)&BoardData.CDRegAddr );
	BoardData.CDRegAddr &= (unsigned long)(~PCI_BASE_ADDRESS_SPACE_IO);

	/* Try to claim our I/O space */
	if (request_region(BoardData.CDRegAddr, BoardData.boardSize,
			   DEVDRVR_ID) == 0) {
		DPRINTK("Unable to allocate adapter region\n");
		rc = -EINVAL;
		goto err_out;
	}

	/* Read the cash drawer config byte */
	cdConfig = inb(BoardData.CDRegAddr + 1);

	DPRINTK("Cash Drawer Config=0x%x\n", cdConfig); 

	/* Cash Drawer 1 Attached (0 = attached, 1 = not) */
	if (!(cdConfig & 0x20)) {
		BoardData.SlotDrawers += 1;
	}

	/* Cash Drawer 2 Attached (0 = attached, 1 = not) */
	if (!(cdConfig & 0x80)) {
		BoardData.SlotDrawers += 1;
	}

	goto out;

err_out:
	DPRINTK("Unload module\n");

	if (unregister_chrdev(aipbcd_major, DEVDRVR_ID) != 0) {
		DPRINTK("Unable to unregister device\n");
	}

out:
	DPRINTK("End Init\n");

	return rc;
}


/* aipbcd_exit: Exit (unload) device driver 
 * ************************************************************************
 *
 *  Function Name:  aipbcd_exit
 *
 *  Purpose:        Unload this device driver (called at rmmod time)
 *
 *  Description:    Release (unmap) memory, I/O space, and IRQ.
 *
 *  Input:          none
 *
 *  Output:         none
 *
 * ************************************************************************
 */

static void __exit aipbcd_exit(void)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 0)
	if (MOD_IN_USE)
	  {
	    printk("%s: device busy, remove delayed\n", DEVDRVR_ID);
	    return;
	  }
#endif

	if (BoardData.CDRegAddr != 0) {
		DPRINTK("Release port (0x%x) regions\n", BoardData.CDRegAddr);
		release_region(BoardData.CDRegAddr, BoardData.boardSize);
	}

	DPRINTK("Unload module\n");

	if (unregister_chrdev(aipbcd_major, DEVDRVR_ID) != 0) {
		printk(KERN_WARNING "%s: Unable to unregister device\n",
		       DEVDRVR_ID);
	} else {
		DPRINTK("Module unloaded\n");
	}    
}


/* aipbcd_open: Open device driver 
 * ************************************************************************
 *
 *  Function Name:  aipbcd_open
 *
 *  Purpose:        Open this device driver
 *
 *  Description:    Setup NVRAM and enable PC interrupts (device channel).
 *
 *  Input:          inode  -   
 *                  file   - 
 *
 *  Output:         0  -- success
 *                  -1 -- error
 *
 * ************************************************************************
 */

static int aipbcd_open(struct inode *iNode, struct file *file)
{  
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 0)
	MOD_INC_USE_COUNT;
#endif

	++BoardData.OpenCount;

	DPRINTK("END OpenCount=%d\n", BoardData.OpenCount);

	return 0;
}


/* aipbcd_release: Close device driver 
 * ************************************************************************
 *
 *  Function Name:  aipbcd_release
 *
 *  Purpose:        Close this device driver
 *
 *  Description:    Disable PC interrupts (device channel).
 *
 *  Input:          inode  -   
 *                  file   - 
 *
 *  Output:         0  -- success
 *                  -1 -- error
 *
 * ************************************************************************
 */

static int aipbcd_release(struct inode *iNode, struct file *file)
{  
	if (--BoardData.OpenCount == 0) {
		BoardData.cd_pid = 0;
	}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 0)
	MOD_DEC_USE_COUNT;
#endif

	DPRINTK("END OpenCount=%d\n", BoardData.OpenCount);

	return 0;
}


/* aipbcd_read: Read from device 
 * ************************************************************************
 *
 *  Function Name:  aipbcd_read
 *
 *  Purpose:        Read from this device
 *
 *  Description:    Not supported by this device.
 *
 *  Input:          file  -   
 *                  buf   - 
 *                  count - 
 *                  ppos  - 
 *
 *  Output:         -EINVAL -- invalid
 *
 * ************************************************************************
 */

static ssize_t aipbcd_read(struct file *file, char *buf, size_t count,
			      loff_t *ppos)
{
	DPRINTK("Error: read unsupported\n");

	return -EINVAL;
}


/* aipbcd_write: Write to device 
 * ************************************************************************
 *
 *  Function Name:  aipbcd_write
 *
 *  Purpose:        Write to this device
 *
 *  Description:    Not supported by this device.
 *
 *  Input:          file  -   
 *                  buf   - 
 *                  count - 
 *                  ppos  - 
 *
 *  Output:         -EINVAL -- invalid
 *
 * ************************************************************************
 */

static ssize_t aipbcd_write(struct file *file, const char *buf, size_t count,
			       loff_t *ppos)
{
	DPRINTK("Error: write unsupported\n");

	return -EINVAL;
}


/* aipbcd_ioctl: I/O Control for this device driver 
 * ************************************************************************
 *
 *  Function Name:  aipbcd_ioctl
 *
 *  Purpose:        Handle ioctl calls to this device driver
 *
 *  Description:    Call correct function based on ioctl request.
 *
 *  Input:          inode  -   
 *                  file   - 
 *                  cmd    - ioctl request
 *                  arg    - pointer to argument structure
 *
 *  Output:         0  -- success
 *                  -1 -- error
 *
 *  Note:           Remember that FIOCLEX, FIONCLEX, FIONBIO, and FIOASYN 
 *                  are reserved ioctl cmd numbers
 *
 * ************************************************************************
 */

static int aipbcd_ioctl(struct inode *inode, struct file *file,
			   unsigned int cmd, unsigned long arg)
{
	int rc = 0;
	int size = _IOC_SIZE(cmd);	 /* the size bitfield in cmd */

	/* Extract the type and number bitfields, and don't decode
	 * wrong cmds;  return EINVAL before verify_area()
	 */
	if (_IOC_TYPE(cmd) != IOCTL_MAGIC) {
		DPRINTK("INCORRECT IOC_TYPE\n");
		rc = -EINVAL;
		goto out;
	}

	if ((_IOC_NR(cmd) < IOCTL_CD_MINNR) || (_IOC_NR(cmd) > IOCTL_CD_MAXNR)) {
		DPRINTK("INVALID IOC_NR\n");
		rc = -EINVAL;
		goto out;
	}

	if (_IOC_DIR(cmd) & _IOC_READ) {
		rc = verify_area(VERIFY_WRITE, (void *) arg, size);
	} else if (_IOC_DIR(cmd) & _IOC_WRITE) {
		rc = verify_area(VERIFY_READ, (void *) arg, size);
	}

	if (rc != 0) {
		goto out;
	}

	switch (_IOC_NR(cmd)) {
	case _IOC_NR(GET_SLOT_INFO):
       		rc = get_slot_info((char *) arg);
		break;
	case _IOC_NR(OPEN_CD):
       		rc = open_cd_drawer((char *) arg);
		break;
	case _IOC_NR(GET_CD_STATUS):
       		rc = get_cd_status((char *) arg);
		break;
	case _IOC_NR(DISABLE_ASYNC_CD_COMM):
       		rc = disable_async_comm((char *) arg);
		break;
	case _IOC_NR(ENABLE_ASYNC_CD_COMM):
		rc = enable_async_comm((char *) arg);
		break;
	case _IOC_NR(SET_CD_INTERFACE):
		rc = set_cd_interface((char *) arg);
		break;
	default:
		rc = -EINVAL;
		break;
	}

out:

	return rc;
} 


/* disable_async_comm: Request to disable notification of cash drawer interrupts
 * ************************************************************************
 *
 *  Function Name:  disable_async_comm
 *
 *  Purpose:        Disable notification of cash drawer interrupts
 *
 *  Description:    This function disables asynchronous communication of 
 *                  cash drawer interrupts to user mode by clearing out 
 *                  the cash drawer process id and cash drawer async_comm 
 *                  flag in the BOARDDATA structure.
 *
 *  Input:          pBuffer - pointer to argument structure
 *
 *  Output:         0       - success
 *                  -EFAULT - error 
 *
 *    On success, CompletionCode will be RPDONE
 *
 * ************************************************************************
 */

static int disable_async_comm(char *buffer)
{
	COMPLETION_CODE_PARMS parms = {0};
	int rc = 0;

	DPRINTK("cd_pid=0x%x\n", BoardData.cd_pid);

	BoardData.cd_pid = 0;

	parms.CompletionCode = RPDONE;

	if (copy_to_user(buffer, &parms, sizeof(parms)) != 0) {
		rc = -EFAULT;
	}

	return rc;
}


/* enable_async_comm: Request to enable notification of cash drawer interrupts
 * ************************************************************************
 *
 *  Function Name:  enable_async_comm
 *
 *  Purpose:        Enable notification of cash drawer interrupts
 *
 *  Description:    This function enables asynchronous communication of 
 *                  cash drawer interrupts to user mode by setting  
 *                  the cash drawer process id and cash drawer async_comm 
 *                  flag in the BOARDDATA structure.
 *
 *  Input:          pBuffer - pointer to argument structure
 *
 *  Output:         0       - success
 *                  -EFAULT - error 
 *
 *    On success, CompletionCode will be RPDONE
 *
 * ************************************************************************
 */

static int enable_async_comm(char *buffer)
{
	ENABLE_ASYNC_COMM_PARMS parms = {0};
	int rc = 0;

	DPRINTK("cd_pid=0x%x\n", BoardData.cd_pid);

	if (BoardData.cd_pid != 0) {
		rc = -EBUSY;
		goto out;
	}

	if (copy_from_user(&parms, buffer, sizeof(parms)) != 0) {
		rc = -EFAULT;
		goto out;
	}

	BoardData.cd_pid = (pid_t) parms.ProcessID;

	parms.CompletionCode = RPDONE;

	DPRINTK("cd_pid=0x%x\n", BoardData.cd_pid);

	if (copy_to_user(buffer, &parms, sizeof(parms))) {
		rc = -EFAULT;
	}
out:

	return rc;
}


/* get_slot_info: Get adapter id and number of cash drawers
 * ************************************************************************
 *
 *  Function Name:  get_slot_info
 *
 *  Purpose:        Get adapter id and number of ports
 *
 *  Description:    This function gets the adapter id and number of ports
 *                  for the adapter in this given slot and returns it to
 *                  the caller in the IOCTL parameter block.
 *
 *  Input:          pBuffer - pointer to argument structure
 *
 *  Output:         0       - success
 *                  -EFAULT - error
 *
 *    On success, CompletionCode will be RPDONE
 *
 * ************************************************************************
 */

static int get_slot_info(char *buffer)
{
	GET_SLOT_INFO_PARMS parms = {0};
	unsigned char cdConfig = 0x00;
	int rc = 0;

	/* Reset SlotDrawers */
	BoardData.SlotDrawers = 0;

	/* Read the cash drawer config byte again since someone could have
	 * attached a cash drawer since load time
	 */
	cdConfig = inb(BoardData.CDRegAddr + 1);

	DPRINTK("Cash Drawer Config=0x%x\n", cdConfig); 

	/* Cash Drawer 1 Attached (0 = attached, 1 = not) */
	if (!(cdConfig & 0x20)) {
		BoardData.SlotDrawers += 1;
		DPRINTK("Cash Drawer 1 Attached\n"); 
	}

	/* Cash Drawer 2 Attached (0 = attached, 1 = not) */
	if (!(cdConfig & 0x80)) {
		BoardData.SlotDrawers += 1;
		DPRINTK("Cash Drawer 2 Attached\n"); 
	}

	parms.AdapterID[1]   = SP500_DEVICE_ID;
	parms.SlotDrawers[1] = BoardData.SlotDrawers;

	DPRINTK("AdapterID=%#x DrawerCount=%d\n", parms.AdapterID[1],
	        parms.SlotDrawers[1]);

	parms.CompletionCode = RPDONE;
	if (copy_to_user(buffer, &parms, sizeof(parms))) {
		rc = -EFAULT;
	}

	DPRINTK("End rc=%d\n", rc);

	return rc;
}

/* get_cd_status: Request to get cash drawer status byte
 * ************************************************************************
 *
 *  Function Name:  get_cd_status
 *
 *  Purpose:        Get the cash drawer status byte
 *
 *  Description:    This function returns to the caller the cash drawer 
 *                  status byte for the specified adapter.
 *
 *  Input:          buffer - pointer to argument structure
 *
 *  Output:         0       - success
 *                  -EFAULT - error 
 *
 *    On success, CompletionCode will be RPDONE
 *
 * ************************************************************************
 */

static int get_cd_status(char *buffer)
{
	GET_CD_STATUS_PARMS parms = {0};
	unsigned char cd_status;
	unsigned char translated_status;
	int rc = 0;

	if (copy_from_user(&parms, buffer, sizeof(parms)) != 0) {
		rc = -EFAULT;
		goto out;
	}

	cd_status = inb(BoardData.CDRegAddr + 1);

	DPRINTK("Raw Cash Drawer Status 0x%x\n", cd_status);

	translated_status = cd_status & 0x0F;
	if (cd_status & 0x80)
		translated_status |= CD2_CONNECTED;
	if (cd_status & 0x40)
		translated_status |= CD2_OPEN;
	if (cd_status & 0x20)
		translated_status |= CD1_CONNECTED;
	if (cd_status & 0x10)
		translated_status |= CD1_OPEN;

	parms.CDStatus = (unsigned long)translated_status;

	DPRINTK("Cash Drawer Status 0x%x\n", parms.CDStatus);

	parms.CompletionCode = RPDONE;

	if (copy_to_user(buffer, &parms, sizeof(parms))) {
		rc = -EFAULT;
	}

out:
	return rc;
}


/* open_cd_drawer: Request to open a cash drawer
 * ************************************************************************
 *
 *  Function Name:  open_cd_drawer
 *
 *  Purpose:        Open the specified cash drawer
 *
 *  Description:    This function opens the cash drawer specified by the
 *                  arguments passed to it. This call has no effect when  
 *                  the cash drawer interface type is legacy (4694).
 *
 *  Input:          pBuffer - pointer to argument structure
 *
 *  Output:         0       - success
 *                  -EFAULT - error 
 *
 *    On success, CompletionCode will be RPDONE
 *
 * ************************************************************************
 */

static int open_cd_drawer(char *buffer)
{
	int index;
	OPEN_CD_PARMS parms = {0};
	int rc = 0;

	if (copy_from_user(&parms, buffer, sizeof(parms)) != 0) {
		rc = -EFAULT;
		goto out;
	}

	for (index = 0; index < OPEN_CMD_SIZE; index++) {
		if ((parms.WhichCD == 1) || ((index != 6) && (index != 7))) {
			DPRINTK("outb(0x%x, 0x%x)\n", OpenCDCmd[index],
				BoardData.CDRegAddr);

			outb(OpenCDCmd[index], (BoardData.CDRegAddr));
		} else if ((index == 6) && (parms.WhichCD == 2)) {
			DPRINTK("outb(0xFD, 0x%x)\n", BoardData.CDRegAddr);

			outb(0xFD, (BoardData.CDRegAddr));
		} else if ((index == 7) && (parms.WhichCD == 2)) {
			DPRINTK("outb(0x7D, 0x%x)\n", BoardData.CDRegAddr);

			outb(0x7D, BoardData.CDRegAddr);
		}
	}

	DPRINTK("Open Cash Drawer %d\n", parms.WhichCD);

	parms.CompletionCode = RPDONE;

	if (copy_to_user(buffer, &parms, sizeof(parms)) != 0) {
		rc = -EFAULT;
	}

out:
	return rc;
}


/* set_cd_interface: Request to set the cash drawer interface
 * ************************************************************************
 *
 *  Function Name:  set_cd_interface
 *
 *  Purpose:        Set the cash drawer interface to either legacy (4694)
 *                  or PCI mode.
 *
 *  Description:    This function sets the cash drawer interface based
 *                  on the values in the arguments passed to it. Both 
 *                  the interface type (4694 or PCI) and the pulse width
 *                  (100 or 200ms) are set.
 *
 *  Input:          pBuffer - pointer to argument structure
 *
 *  Output:         0       - success
 *                  -EFAULT - error 
 *
 *    On success, CompletionCode will be RPDONE
 *
 * ************************************************************************
 */

static int set_cd_interface(char *buffer)
{
	SET_CD_INTERFACE_PARMS parms = {0};
	int rc = 0;

	DPRINTK("Interface set\n");

	if (copy_from_user(&parms, buffer, sizeof(parms)) != 0) {
		rc = -EFAULT;
		goto out;
	}

	parms.CompletionCode = RPDONE;

	if (copy_to_user(buffer, &parms, sizeof(parms)) != 0) {
		rc = -EFAULT;
	}

out:
	return rc;
}

