/* 
 * Module Name:     aipmtn-2.6.c
 *
 * Description:     Motion Sensor Driver for Linux 2.6.X
 *
 *  (C) Copyright 2004,2005 International Business Machines Corporation.
 *  All right reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1.  Redistributions of source code must retain the above copyright notice,
 *     this list of conditions and the following disclaimer.
 *
 * 2.  Redistributions in binary form must reproduce the above copyright notice,
 *     this list of conditions and the following disclaimer in the documentation
 *     and/or other materials provided with the distribution.
 *
 * 3.  The name of the author may not be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *
 * Should you need to contact me, the author, you can do so either by
 * e-mail - mail your message to <nobody@us.ibm.com>.
 */

/* System Headers */
#include <linux/version.h>
#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h> 
#include <linux/compiler.h>
#include <linux/pci.h>
#include <linux/init.h>
#include <linux/ioport.h>
#include <linux/delay.h>
#if LINUX_VERSION_CODE > KERNEL_VERSION(2, 5, 0)
#include <linux/moduleparam.h>
#endif
#include <linux/timer.h>
#include <linux/interrupt.h>
#include <asm/uaccess.h>
#include <linux/fs.h>


/* Local configuration constants */
#ifndef __LINUX__
#define __LINUX__
#endif
#ifndef __UNIX__
#define __UNIX__
#endif
#include "aipkernel.h"


MODULE_AUTHOR("IBM Retail Store Solutions");
MODULE_DESCRIPTION("IBM Motion Sensor Driver");
MODULE_LICENSE("GPL compatible");


/* Function Prototypes */
static int __init aipmtn_init(void);
static void __exit aipmtn_exit (void);
static int aipmtn_open(struct inode *inode, struct file *file);
static ssize_t aipmtn_read(struct file *file, char *buf, size_t count,
			   loff_t *ppos);
static ssize_t aipmtn_write(struct file *file, const char *buf, size_t count,
			    loff_t *ppos);
static int aipmtn_release(struct inode *inode, struct file *file);
static int aipmtn_ioctl(struct inode *inode, struct file *file,
			unsigned int cmd, unsigned long arg);

static int aipmtn_get_slot_info(char *pBuffer);
static int aipget_motion_status(char *pBuffer);

module_init(aipmtn_init);
module_exit(aipmtn_exit);

/* Static Global Variables */
#define ADAPTER_ID "aipmtn"
#define DEFAULT_MAJOR_DEV 249

int aipmtn_major = DEFAULT_MAJOR_DEV;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 0)
MODULE_PARM(aipmtn_major, "i");
#else
module_param_named(major, aipmtn_major, int, 0644);
#endif


struct pci_dev *pcimtn = NULL;

#ifdef DEBUG
#define DPRINTK(fmt, args...) printk(KERN_DEBUG "%s(%s): " fmt, ADAPTER_ID, __FUNCTION__ , ## args)
#else
#define DPRINTK(fmt, args...)
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 4, 23)
typedef void irqreturn_t;
/* #define IRQ_NONE */
/* #define IRQ_HANDLED */
#define IRQ_RETVAL(arg)
#endif

/* Information for IBM PCI ASIC chip (Vendor ID = 0x1014, Device ID = 0x02A2) */
#define RW_VENDOR_ID   	0x1014
#define RW_DEVICE_ID	0x02A2
#define RW_SUBSYSTEM_ID	0x02A3
#define RW_SUBVENDOR_ID	0x1014

/* This value is taken from the EEPROM General storage. It is stored at
 * address 0x38
 */
#define OFFSET_TO_MACHINE_MODEL 0x38
#define RW_MODEL		0x30343834
#define KIOSK_MODEL	       	0x35333834


/* This value is taken from the EEPROM General storage. It is stored at
 * address 0x3C. The correct reading should be 563.  This value is in ASCII
 * mode and should be 353633.  Since its in endian format it is read backward.
 */
#define OFFSET_TO_MACHINE_TYPE	0x3C
#define RW_MTN_SENSOR_MACHINE	0x333635
#define RW_MTN_SENSOR_PRELOAD   0x333657
#define RW_MTN_SENSOR_4840_5A3  0x334135
#define RW_MTN_SENSOR_4840_WA3  0x334157
#define RW_MTN_SENSOR_4840_E63  0x333645
#define RW_MTN_SENSOR_4840_573  0x333735
#define RW_MTN_SENSOR_4840_W73  0x333757

#define RW_MTN_SENSOR_KIOSK     0x333531
#define RW_MTN_SENSOR_4835_15W  0x573531
#define RW_MTN_SENSOR_4835_E53  0x333545

#define CONTROL_REG		0x0
#define INTERRUPT_STATUS_REG   	0x01
#define PCI_BOARD_SIZE		0x10	/* 16 bytes */


#define FALSE 0
#define TRUE  1

static unsigned long function0port = 0;
static unsigned long mtn_state = 0;
static unsigned short irq_dev_id = 0;


#define INTEL_LPC_VENDOR_ID	 0x8086		/*Intel */
#define INTEL_LPC_DEVICE_ID	 0x24C0  	/* LPC-Interface */
#define PORT_ADDRESS_OFFSET_BYTE 64   		/* 40h in configuration space */
#define	PORT_OFFSET		 0xFFFFFF2B 	/* Offset Address */

static int              get_port_address;
static int              RW = 0;
static int              port_address;
      

/* aipmtn_hw_int_handler: Handler for interrupt event
 * ***********************************************************************
 *
 * Function Name:   aipmtn_hw_int_handler
 *
 * Purpose:         Handler for interrupt event
 *
 * Description:     When an interrupt occurs
 *
 * Input:
 *
 * Output:          None
 *
 * ***********************************************************************
 */
static irqreturn_t aipmtn_hw_int_handler(int irq, void *dev_id,
					    struct pt_regs *regs)
{
	unsigned char func_reg = inb(function0port + INTERRUPT_STATUS_REG);
	int rc = FALSE;

	if (func_reg & 0x01) {
		DPRINTK("INTERRUPT ASSERTED\n");

		/* If interrupt, than clear bit 0 of function0 Interrupt Status
		 * Register 
		 */
		outb(0x00, function0port + INTERRUPT_STATUS_REG);
		rc = TRUE;
	}

	return IRQ_RETVAL(rc == TRUE ? IRQ_HANDLED : IRQ_NONE);
}


struct file_operations aipmtn_fops = {
	.owner      = THIS_MODULE,
	.read       = aipmtn_read,
	.write      = aipmtn_write,
	.ioctl      = aipmtn_ioctl,
	.open       = aipmtn_open,
	.release    = aipmtn_release,
};


/* aipmtn_init: Loads and Initializes the driver
 * **********************************************************************
 *                                                                      *
 *  Loads the aipmtn driver.  This is called when "insmod aipmtn.o is   *
 *  invoked on the command line.                                        *
 *                                                                      *
 * **********************************************************************
 */
static int __init aipmtn_init(void)
{  
	int err = 0;
	int rc  = 0;
	unsigned char *pEEROMConfig = 0;
	unsigned long EEROMConfigAddr = 0;
	unsigned long EEROMConfigLength = 0;
	unsigned long machineModel;
	unsigned long machineType;
	u32 ctrl_register = 0; 

	/* Register as a device with kernel. */
	if ((err = register_chrdev(aipmtn_major, ADAPTER_ID, &aipmtn_fops))) {
		printk(KERN_ERR "%s: Failure to register device (error=%d)\n",
		       ADAPTER_ID, -err);               
		goto out;
	}

	DPRINTK("Device registered with the kernel successfully.\n");   

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 9)
	pcimtn = pci_find_device(RW_VENDOR_ID, RW_DEVICE_ID, pcimtn);
#else
	pcimtn = pci_get_device(RW_VENDOR_ID, RW_DEVICE_ID, pcimtn);
#endif

	if (pcimtn != NULL) {
		DPRINTK("Adapter found...\n");

		if ((rc = pci_enable_device(pcimtn))) {
			printk(KERN_ERR "%s: Cannot enable PCI device\n",
			       ADAPTER_ID);
			goto err_out;
		}

		DPRINTK("PCI Device is Enabled\n");

		EEROMConfigAddr = pci_resource_start(pcimtn, 1);
		EEROMConfigLength = pci_resource_len(pcimtn, 1);

		pEEROMConfig = (unsigned char *) ioremap(EEROMConfigAddr,
							 EEROMConfigLength);
		if (pEEROMConfig == 0) {
			goto err_out;
		}

		machineModel = readl((pEEROMConfig + OFFSET_TO_MACHINE_MODEL));
		machineType  = readl((pEEROMConfig + OFFSET_TO_MACHINE_TYPE));
		machineType  = (machineType & 0x00FFFFFF);

		iounmap((unsigned char *) pEEROMConfig);

		DPRINTK("Model=0x%lx, Type=0x%lx\n", machineModel, machineType);

		if (machineModel == RW_MODEL &&
		    (machineType == RW_MTN_SENSOR_MACHINE ||
		     machineType == RW_MTN_SENSOR_PRELOAD ||
		     machineType == RW_MTN_SENSOR_4840_5A3 ||
		     machineType == RW_MTN_SENSOR_4840_WA3 ||
		     machineType == RW_MTN_SENSOR_4840_E63 ||
		     machineType == RW_MTN_SENSOR_4840_573 ||
		     machineType == RW_MTN_SENSOR_4840_W73 )) {
			mtn_state = 1;
			DPRINTK("RW_MODEL found, Type=0x%lx\n", machineType);
		} else if (machineModel == KIOSK_MODEL &&
			   (machineType == RW_MTN_SENSOR_KIOSK ||
			    machineType == RW_MTN_SENSOR_4835_15W ||
			    machineType == RW_MTN_SENSOR_4835_E53 )) {
			mtn_state = 1;
			DPRINTK("KIOSK_MODEL found, Type=0x%lx\n", machineType);
		} else {
			DPRINTK("UNSUPPORTED MODEL Found\n");               
			goto err_out;
		}

		/* should always use irq number from pci_dev routine 
		 * because of differences in the PCI_INTERRUPT_LINE macro
		 */
		rc = request_irq(pcimtn->irq, aipmtn_hw_int_handler,
				 SA_INTERRUPT | SA_SHIRQ, ADAPTER_ID,
				 &irq_dev_id);

		DPRINTK("request_irq rc=%d\n", rc);

		if (rc != 0) {
			DPRINTK("ERROR Requesting irq\n");

			goto err_out;
		}
		DPRINTK("Hooked Interrupt Handler\n");

		if (pci_read_config_dword(pcimtn, PCI_BASE_ADDRESS_0, 
					  (u32 *)&function0port ) < 0) {
			DPRINTK("Unable to read BAR 0\n");

			goto err_out;
		}

		function0port &= (unsigned long)(~PCI_BASE_ADDRESS_SPACE_IO);
		/* Request I/O Space for BAR 0 size 16 bytes */
		if (request_region(function0port, PCI_BOARD_SIZE, 
				   ADAPTER_ID) == NULL) {
			DPRINTK("Unable to request region\n");
			goto err_out;
		}

		DPRINTK("region requested\n");

		/* Reading first byte of BAR 0 */
		ctrl_register = inb(function0port + CONTROL_REG); 

		DPRINTK("presence_sensor_ctrl_register (before)=0x%x\n", 
			 ctrl_register);

		/* enable bit 2 */
		ctrl_register |= 0x04;          

		/* write to I/O Space */
		outb(ctrl_register, function0port);

		/* Reading first and second byte to verify results */
		DPRINTK("presence_sensor_ctrl_register (after)=0x%x\n", 
			 inb(function0port + CONTROL_REG));
		DPRINTK("function0_interrupt_status_register=0x%x\n",
			 inb(function0port + INTERRUPT_STATUS_REG));

		RW = 1;

		rc = 0;
	} else {

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 9)
		pcimtn = pci_find_device(INTEL_LPC_VENDOR_ID,
					 INTEL_LPC_DEVICE_ID, pcimtn);
#else
		pcimtn = pci_get_device(INTEL_LPC_VENDOR_ID,
					INTEL_LPC_DEVICE_ID, pcimtn);
#endif

		if (pcimtn == NULL) {
			printk(KERN_ERR "%s: Non-supported machine\n",
			       ADAPTER_ID);                     
			goto err_out;
		}

		/* Read the Base Port Address Value. */
		if ((pci_read_config_dword(pcimtn, PORT_ADDRESS_OFFSET_BYTE,
					   (u32 *)&get_port_address)) < 0) {
			DPRINTK("Cannot read PCI space\n");
			goto err_out;
		}

		mtn_state = 1;

		/* Read Value at Base Port Address and Offset it. */
		port_address = get_port_address & 0xFFFE;
		port_address = port_address + 0x2b;

		DPRINTK("Port Address is %x\n", get_port_address);
		rc = 0;       
	}         
	DPRINTK("RW Type is %d\n", RW);
	goto out;

err_out:
	/* Unregister device if it is not equal to the correct model number */
	mtn_state = 0;

	if (unregister_chrdev(aipmtn_major, ADAPTER_ID) != 0) {
		printk(KERN_ERR "%s: Unable to unregister device\n", ADAPTER_ID);
	}
	rc = -EINVAL;

out:     
	return rc;
}


/* aipmtn_exit: Exits the driver once rmmod is called 
 * ***********************************************************************
 *
 * Function Name:   aipmtn_exit
 *
 * Purpose:         Exits the driver once rmmod is called
 *
 * Description:     rmmod of device
 *
 * Input:
 *
 * Output:          None
 *
 * ***********************************************************************
 */
static void __exit aipmtn_exit(void)
{

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 0)
	if (MOD_IN_USE) {
		DPRINTK("device busy, remove delayed\n");
		return;
	}
#endif

	release_region(function0port, PCI_BOARD_SIZE);
	if (pcimtn->irq != 0) {
		free_irq(pcimtn->irq, &irq_dev_id);
	}

	if (unregister_chrdev(aipmtn_major, ADAPTER_ID) != 0) {
		printk(KERN_ERR "%s:unregister device failed\n", ADAPTER_ID);
	}

	DPRINTK("Module removed\n");
}

/* aipmtn_open: Exits the driver once rmmod is called 
 * ***********************************************************************
 *
 * Function Name:   aipmtn_open
 *
 * Purpose:         Activites properties and other initialization not performed
 *                  during the init
 *
 * Description:     on device open
 *
 * Input:
 *
 * Output:          integer value
 *
 * ***********************************************************************
 */
static int aipmtn_open(struct inode *inode, struct file *flip)
{
	DPRINTK("open()\n");

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 0)
	MOD_INC_USE_COUNT;
#endif

	return 0;
}


/* aipmtn_read:
 * ***********************************************************************
 *
 * Function Name:   aipmtn_read
 *
 * Purpose:         no operation
 *
 * Description:     on device read
 *
 * Input:
 *
 * Output:          integer value
 *
 * ***********************************************************************
 */
static int aipmtn_read(struct file *flip, char *buf, size_t count, 
			  loff_t *f_pos)
{
	DPRINTK("Error: read unsupported\n");

	return -EINVAL;
}


/* aipmtn_write:
 * ***********************************************************************
 *
 * Function Name:   aipmtn_write
 *
 * Purpose:         no operation
 *
 * Description:     on device write
 *
 * Input:
 *
 * Output:          integer value
 *
 * ***********************************************************************
 */
static int aipmtn_write(struct file *flip, const char *buf, size_t count, 
			   loff_t *f_pos)
{ 
	DPRINTK("Error: write unsupported\n");

	return -EINVAL;
}


/* aipmtn_release:
 * **********************************************************************
 *
 * Function Name:   aipmtn_release
 *
 * Purpose:         Releases properties and other initialization 
 *
 * Description:     on device release
 *
 * Input:
 *
 * Output:          integer value
 *
 * ***********************************************************************
 */
static int aipmtn_release(struct inode *inode, struct file *flip)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 0)
	MOD_DEC_USE_COUNT;
#endif

	DPRINTK("END close()\n");
	return 0;
}


/* aipmtn_ioctl:
 * **********************************************************************
 *
 * Function Name:   aipmtn_ioctl
 * 
 * Purpose:         ioctl service handler
 *
 * Description:     on device ioctl call
 *
 * Input:
 *
 * Output:          integer value
 *
 * ***********************************************************************
 */
static int aipmtn_ioctl(struct inode *inode, struct file *flip, 
			   unsigned int cmd, unsigned long arg)
{
	int rc = 0;     
	switch (_IOC_NR(cmd)) {
	case _IOC_NR(GET_MTN_STATUS):
		rc = aipget_motion_status((char *) arg);
		break;  

	case _IOC_NR(GET_SLOT_INFO):
		rc = aipmtn_get_slot_info((char *) arg);
		break;

	default:
		rc = -EINVAL;
		break;
	}
	return rc;
}


/* aipmtn_get_slot_info:
 * **********************************************************************
 *
 * Function Name:   aipmtn_get_slot_info
 * 
 * Purpose:         obtain slot information
 *
 * Description:     on device ioctl call
 *
 * Input:
 *
 * Output:          integer value
 *
 * ***********************************************************************
 */
static int aipmtn_get_slot_info(char *pBuffer)
{
	int rc = 0;

	GET_SLOT_INFO_PARMS gsip = {0};

	DPRINTK("aipmtn_get_slot_info\n");

	gsip.SlotType[1] = 1;
	gsip.AdapterID[1] = pcimtn->device;
	gsip.SlotMotionSensor[1] = mtn_state;
	gsip.CompletionCode = 0x100;

	if (copy_to_user(pBuffer, &gsip, sizeof(gsip))) {
		DPRINTK("Error Copy GET_SLOT_INFO_PARMS to user space.\n");
		rc = -EFAULT;
	}

	return rc;
}


/* aipget_motion_status
 * **********************************************************************
 *
 * Function Name:   aipget_motion_status
 * 
 * Purpose:         motion status
 *
 * Description:     on device ioctl call
 *
 * Input:
 *
 * Output:          integer value
 *
 * ***********************************************************************
 */
static int aipget_motion_status(char *pBuffer)
{
	int rc = 0;
	unsigned char status;
	GET_MTN_STATUS_PARMS gmsp = {0};        
	gmsp.SlotNumber = 1;

	if (RW) {
		status = inb(function0port);

		DPRINTK("Status of Sensor=%x\n", status);

		status &= 0x01;

		if (status == 0x01) {
			DPRINTK("Motion NOT detected\n");
		} else {
			DPRINTK("Motion detected\n");
		}   
	} else {
		/* We need to read port(LPC PCI Interface Bridge- 42B, bit 4) */

		/* The following write is to clear the Status bit to get
		 * the correct latest status.
		 */
		outb(0x10, port_address);

		/* Read the latest status. */
		status = inb(port_address);                             

		DPRINTK("Status of Sensor=%x\n", status);

		/* Bit4 is Presence Sensor status bit. */
		status &= 0x10;

		if (status == 0x10) {
			/* Return 0 if presence detected */
			status = 0;                         
			DPRINTK("Motion detected\n");
		} else {
			/* Return 1 if presence not detected */
			status = 1;                         
			DPRINTK("Motion NOT detected\n");
		}

	}       
	gmsp.mtnStatus = status;
	gmsp.CompletionCode = 0x100;

	if (copy_to_user(pBuffer, &gmsp, sizeof(gmsp))) {
		DPRINTK("Error Copy GET_MTN_STATUS_PARMS to user space\n");
		rc = -EFAULT;
	}

	return rc;
}
