/*
 * Module Name:     aipikbps.c
 *
 * Description:     PS/2 attached keyboard driver for Linux.
 *
 * (C) Copyright 2004,2005, International Business Machines Corporation.
 * All right reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1.  Redistributions of source code must retain the above copyright notice,
 *     this list of conditions and the following disclaimer.
 *
 * 2.  Redistributions in binary form must reproduce the above copyright notice,
 *     this list of conditions and the following disclaimer in the documentation
 *     and/or other materials provided with the distribution.
 *
 * 3.  The name of the author may not be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Should you need to contact me, the author, you can do so either by
 * e-mail - mail your message to <nobody@us.ibm.com>.
 */

#include <linux/version.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/fs.h>
#include <linux/slab.h>
#include <linux/vmalloc.h>
#include <linux/mm.h>
#include <linux/ioctl.h>
#include <linux/delay.h>
#include <linux/sched.h>
#include <linux/kbd_kern.h>
#include <linux/fs.h>
#include <asm/delay.h>
#include <asm/segment.h>
#include <asm/io.h>
#include <asm/segment.h>
#include <asm/uaccess.h>
#include <asm/system.h>

MODULE_LICENSE("GPL compatible");

/* ****************************************************************
 * local include files
 * ****************************************************************
 */
#define CANPOS


#include "aipikbps.h"
#include "aipikpssc.h"

/* ****************************************************************
 * constant definitions
 * ****************************************************************
 */
#define SUCCESS             0
#define IBM_POS_KBD_MAJOR   246
#define POS_KBD_NODE_NAME   "aipikbps"
#define CMD_WAIT_TIMEOUT    HZ*50;

// #define KBD_RESET           0


#ifdef DEBUG
#define DPRINTK(fmt, args...) printk(KERN_DEBUG "%s(%s): " fmt, POS_KBD_NODE_NAME, __FUNCTION__ , ## args)
#else
#define DPRINTK(fmt, args...)
#endif

#define EPRINTK(fmt, args...) printk(KERN_ERR "%s(%s): " fmt, POS_KBD_NODE_NAME, __FUNCTION__ , ## args)
#define IPRINTK(fmt, args...) printk(KERN_INFO "%s(%s): " fmt, POS_KBD_NODE_NAME, __FUNCTION__ , ## args)
#define WPRINTK(fmt, args...) printk(KERN_WARNING "%s(%s): " fmt, POS_KBD_NODE_NAME, __FUNCTION__ , ## args)

/* ***************************************************************
 * Function Prototypes
 * ***************************************************************
 */
static int kb_open(struct inode *inode, struct file *file);
static int kb_release(struct inode *inode, struct file *file);
static int kb_read(struct file *file, char *buf, size_t count, loff_t *ppos);
static int kb_write(struct file *file, const char *buf, size_t count,
		    loff_t *ppos);
static int kb_ioctl(struct inode *inode, struct file *fp, unsigned int cmd,
		    unsigned long ioctl_arg);
static int pos_scancode_filter(unsigned char *pScancode);

#if defined(KBD_RESET)
static int PosKbdReset(void);
#endif 

static int PosKbdSendData(unsigned char *buf, unsigned char size);
static int PosKbdReqDevInfo(void);
static int PosMsrReqDevInfo(void);
static unsigned char CollectingKBDStatus(unsigned char scancode);
static unsigned char CollectingMSRStatus(unsigned char scancode);
static unsigned char DoubleKeyProcessing(unsigned char key,
					 unsigned char fBreak);
static void SetShiftKeyState(unsigned char scanCode,
			     POS_KEYBOARD_SCAN_STATE *scanState);
static void set_leds_default_state(void);
static void pos_process_code_update(void);
static int  SetDefaultConfiguration(void);
static unsigned char CollectingKBDId(unsigned char scancode);
static int  PosKbdReadId(void);

static unsigned char collecting_canpos_write_config_ack(unsigned char sc);
static unsigned char colleting_canpos_read_config_data(unsigned char sc);
static int CanposPosKbdSendData(unsigned char *buf, unsigned int size);
static unsigned char collecting_canpos_normal_ack(unsigned char sc);
static unsigned char collecting_canpos_flash_ack(unsigned char sc);
static void wake_up_canpos_process(void);
static unsigned char check_canpos_key_105_106(unsigned char scancode);


#ifdef CANPOS
static unsigned char collecting_canpos_ver_info(unsigned char sc);
#endif

struct pc_keyb_filter {
	int (*filter_scancode)(unsigned char *scancode);
	int (*write_data)(unsigned char *data, unsigned int length,
			  unsigned int *xferred, unsigned int retries,
			  unsigned int timeout);
};
int  register_pc_keyb_filter(struct pc_keyb_filter *filter);
void unregister_pc_keyb_filter(struct pc_keyb_filter *filter);

/* ***************************************************************
 * Globals
 * ***************************************************************
 */
static unsigned char updateBuffer[MAX_BUFFER_SIZE];
static int           cmd_busy = 0;
static int           aipikbps_MajorNumber = IBM_POS_KBD_MAJOR;
static int           asyncPID = 0;
static int           asyncEnabled = 0;
static unsigned char canpos_key_105_106;
static unsigned char *flashData;

/* ***************************************************************
 * Pass the scancode filter function name.  Upon return ,
 * pos_filter.write_data() will contain the address of the
 * actual write funtion supported by pc_keyb driver.
 * ***************************************************************
 */
static struct pc_keyb_filter pos_filter =
{
	filter_scancode:  pos_scancode_filter,
};

/* ***************************************************************
 * file I/O
 * ***************************************************************
 */
struct file_operations kb_fops = {
	.read = kb_read,
	.write = kb_write,
	.ioctl = kb_ioctl,
	.open = kb_open,
	.release = kb_release, 
};

DECLARE_WAIT_QUEUE_HEAD(wq);
MODULE_PARM(aipikbps_MajorNumber, "i");


/* :H1.kb_init() - Driver entry point.
 * ***********************************************************************
 *
 * Function Name:  kb_init()
 *
 * Purpose:        Intialization routine
 *
 * Description:    entry point for this driver.
 *
 *
 *
 * Output:
 *                 0        - success
 *                 non-zero - failure.
 *
 * Notes:
 *
 * ***********************************************************************
 */

static int __init kb_init(void)
{
	int rc = SUCCESS;
	int cmd_retries = 0;

	/* Initialize keyboard data structure. */
	memset(&posKbdData, 0, sizeof(POS_KEYBOARD_DATA));

	pPosKbdData = &posKbdData;
	memset(&KBDdevinfo, 0, sizeof(DEVICE_INFO));
	memset(&MSRdevinfo, 0, sizeof(DEVICE_INFO));

	/* Register with low level driver, pc_keyb driver. Upon registration,
	 * the pc_keyb driver provides a pointer to write function.  In addtion,
	 * it calls our interrupt routine (pos_scancode_filter()) each time
	 * keyboard interrupts.
	 */
	DPRINTK("register with keyboar driver - atkbd\n");

	/* register_pc_keyb_filter() & unregister_pc_keyb_filter()
	 * todo - 
	 * this should be OK if the patch for atkbd is applied. 
	 * However, we need to find a way to check for the existing of
	 * the symbols before calling them.
	 */
	register_pc_keyb_filter(&pos_filter);

	set_current_state(TASK_INTERRUPTIBLE);
	schedule_timeout(10);

	/*  Read Keyboard ID.  If IBM POS Keyboard, continue with Initialization.
	 */
	rc = PosKbdReadId();
	if (rc != 0) {
		EPRINTK("PosKbdReadId failed rc=%02x\n", rc);

		/* unregister with pc_keyb driver. */
		DPRINTK("unregister with atkbd\n");

		unregister_pc_keyb_filter(&pos_filter);
		return -ENODEV;
	}

	/* Wait until we have received keyboard ID bytes
	 * max wait = 5*10(ms)*20 = 1000 ms, timer tick = 10ms
	 */
	cmd_retries = 20;
	while (cmd_busy && cmd_retries) {
		DPRINTK("cmd_busy = %02x\n", cmd_busy);

		set_current_state(TASK_INTERRUPTIBLE);
		schedule_timeout(5);
		--cmd_retries;

		DPRINTK("cmd_busy=%02x\n", cmd_busy);
	}

	if ((pPosKbdData->PosKbdId.KBDId[0] == 0xAB) && 
	     (pPosKbdData->PosKbdId.KBDId[1] == 0x87)) {
		/* Found IBM POS Keyboard */
		DPRINTK("IBM POS Keyboard Id=0x%x.0x%x\n",
			pPosKbdData->PosKbdId.KBDId[0],
			pPosKbdData->PosKbdId.KBDId[1]);
	} else if ((pPosKbdData->PosKbdId.KBDId[0] == 0xAB) && 
		    (pPosKbdData->PosKbdId.KBDId[1] == 0x41)) {
		/* Found IBM CANPOS Keyboard */
		DPRINTK("Found IBM CANPOS Keyboard Id=0x%x.0x%x\n",
			pPosKbdData->PosKbdId.KBDId[0],
			pPosKbdData->PosKbdId.KBDId[1]);

		CanposKeyboard = 1;
	} else {
		/* Did not find IBM POS Keyboard */
		EPRINTK("No IBM POS Keyboard Id=0x%x.0x%x\n",
			pPosKbdData->PosKbdId.KBDId[0],
			pPosKbdData->PosKbdId.KBDId[1]);

		/* unregister with pc_keyb driver. */
		unregister_pc_keyb_filter(&pos_filter);

		return -ENODEV;
	}

#if defined(KBD_RESET)
	/*  Reset the keyboard.  There is no need to do any reset.
	 *  Otherwise, we loose initial settings.
	 */
	rc = PosKbdReset();
	if (rc != 0) {
		unregister_pc_keyb_filter(&pos_filter);

		return -ENODEV;
	}

	/* sleep for 200 ms since keyboard reset takes time. */
	set_current_state(TASK_UNINTERRUPTIBLE);
	schedule_timeout(500);
#endif 

	/* Request keyboard device information */
	rc = PosKbdReqDevInfo();
	if (rc != 0) {
		EPRINTK("PosKbdDevInfo() failed rc=0x%x", rc);

		/* unregister with pc_keyb driver. */
		unregister_pc_keyb_filter(&pos_filter);

		return -ENODEV;
	}

	/*  Request msr device information */
	rc = PosMsrReqDevInfo();
	if (rc != 0) {
		EPRINTK("PosMsrDevInfo() failed rc=0x%x", rc);

		/* unregister with pc_keyb driver. */
		unregister_pc_keyb_filter(&pos_filter);

		return -ENODEV;
	}

	/* Set default led state i.e turn off all leds. */
	set_leds_default_state();

	/* Initialze Keyboard */
	pPosKbdData->KbdConfigureCmd[0] = KBD_CONFIGURE_CMD;
	pPosKbdData->KbdConfigureCmd[1] = 0x03;
	pPosKbdData->KbdConfigureCmd[2] = 0x00;
	pPosKbdData->KbdConfigureCmd[3] = 0x00;

	pPosKbdData->MsrConfigureCmd[0] = MSR_CONFIGURE_CMD;
	pPosKbdData->MsrConfigureCmd[1] = 0x00;

	/* Send default Configuration i.e. turn on typematic */
	SetDefaultConfiguration();

	/* Register driver */
	rc = register_chrdev(aipikbps_MajorNumber,  /*  major number */
			     POS_KBD_NODE_NAME,	    /*  node name    */
			     &kb_fops);		    /*  file ops     */

	if (rc) {
		EPRINTK("register devnode failed rc=0x%x", rc);
		return -ENODEV;
	} else {
		DPRINTK("devnode created MajorNumber %d", aipikbps_MajorNumber);
	}


	IPRINTK("Loaded, MajorNumber=%d", aipikbps_MajorNumber);

	return rc;
}


/* :H1.kb_exit() - Driver exit.
 * ***********************************************************************
 *
 * Function Name:  kb_exit()
 *
 * Purpose:        driver exit routine.
 *
 * Description:    This routine gets called when driver unloads.
 *
 *
 * Output:
 *                 none
 *
 *
 * Notes:
 *
 * ***********************************************************************
 */

static void __exit kb_exit(void)
{
	int rc = SUCCESS;

	/* unregister driver */
	rc = unregister_chrdev(aipikbps_MajorNumber, POS_KBD_NODE_NAME);
	if (rc != 0)
		EPRINTK("unregister devnode failed rc=0x%x\n", rc);
	else {
		DPRINTK("unregister devnode success\n");
	}

	/* unregister with pc_keyb driver. */
	unregister_pc_keyb_filter(&pos_filter);
}


/* :H1.kb_open()
 * ***********************************************************************
 *
 * Function Name:  kb_open()
 *
 * Purpose:        driver open routine.
 *
 * Description:    This routine gets called when application opens this
 *                 driver.
 *
 *
 * Output:
 *                 0  -  success
 *
 *
 * Notes:
 *
 * ***********************************************************************
 */

static int kb_open(struct inode *inode, struct file *file)
{
	DPRINTK("MAJOR=%d, MINOR=%d\n",
		MAJOR (inode->i_rdev), MINOR (inode->i_rdev));

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 0)
	MOD_INC_USE_COUNT;
#endif
	return SUCCESS;
}


/* :H1.kb_release()
 * ***********************************************************************
 *
 * Function Name:  kb_release()
 *
 * Purpose:        driver close routine.
 *
 * Description:    This routine gets called when application closes this
 *                 driver.
 *
 * Output:
 *                 0  -  success
 *
 * Notes:
 *
 * ***********************************************************************
 */

static int kb_release (struct inode *inode, struct file *file)
{
	DPRINTK("closing character device: %s\n", POS_KBD_NODE_NAME);

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 0)
	MOD_DEC_USE_COUNT;
#endif

	return SUCCESS;
}


/* :H1.kb_read()
 * ***********************************************************************
 *
 * Function Name:  kb_read()
 *
 * Purpose:        driver read routine.
 *
 * Description:    This routine gets called when application sends a
 *                 read request.
 *
 * Output:
 *                 EINVAL
 *
 * Notes:          read request is not supported.
 *
 * ***********************************************************************
 */

static int kb_read(struct file *file, char *buf, size_t count, loff_t *ppos)
{
	DPRINTK("character device: %s:\n\n", POS_KBD_NODE_NAME);

	return -EINVAL;
}


/* :H1.kb_write()
 * ***********************************************************************
 *
 * Function Name:  kb_wirte()
 *
 * Purpose:        driver write routine.
 *
 * Description:    This routine gets called when application sends a
 *                 write request.
 *
 * Output:
 *                 EINVAL
 *
 * Notes:          write request is not supported.
 *
 * ***********************************************************************
 */

static int kb_write(struct file *file, const char *buf, size_t count,
		    loff_t *ppos)
{
	DPRINTK("character device: %s\n", POS_KBD_NODE_NAME);

	return -EINVAL;
}


/* :H1.kb_ioctl()
 * ***********************************************************************
 *
 * Function Name:  kb_ioctl()
 *
 * Purpose:        driver ioctl routine.
 *
 * Description:    This routine gets called when application sends a
 *                 ioctl request.
 *
 * Output:
 *                 SUCCESS, EINVAL
 *
 * Notes:
 *
 * ***********************************************************************
 */

static int kb_ioctl(struct inode *inode, struct file *fp, unsigned int cmd,
		    unsigned long userBuffer)
{
	int                   rc = SUCCESS;
	unsigned char        *pInputBuffer;
	unsigned char        *pBuffer;
	QUERY_KEYBOARD_PARMS  tmpQueryKeyboard;

	KBD_SET_INDICATORS    setIndicators;
	KBD_SET_TONE          setTone;
	unsigned char         setConfig[KBD_CONFIGURE_CMD_LENGTH];
	unsigned char         bytesToSend;
	unsigned char         setTracks[MSR_CONFIGURE_CMD_LENGTH];
	PDOUBLE_KEYS          pParms;
	unsigned char        *pCurUpdate;

	unsigned char        *pNextUpdate;
	unsigned short        CurLength;
	int                   i,j;
	CODE_UPDATE_STR  tmpQueryUpdate;

	unsigned long         CompletionCode;
	unsigned long         Length = 0;

	int                   readConfigErr = 0;
	int                   writeConfigErr = 0;
	int                   flashCmdErr = 0;
	int                   flashErr;
	PPOS_CANPOS_READ_CONFIG      pCanposReadConfig;
	PPOS_CANPOS_WRITE_CONFIG     pCanposWriteConfig;
	PPOS_CANPOS_FLASH_UPDATE     pCanposFlash;
    
  
    MODQUERY_KEYBOARD_PARMS  tmpCanposQueryKeyboard;
    

	ANPOS_COMPLETION_CODE_PARMS completionCodeParms;

	pInputBuffer = (unsigned char *) userBuffer;

	switch (cmd) {
	
	case FN_CANPOS_READ_CONFIG:
	{
		char Canpos_read_config_cmd[] = {0xE4, 0xC1, 0xC2, 0xC3};

		DPRINTK("enter FN_CANPOS_READ_CONFIG\n");

		if (CanposKeyboard != 1) {
			/* invalid command */
			CompletionCode = RPERR;
			copy_to_user(&((POS_CANPOS_READ_CONFIG *) userBuffer)->CompletionCode,
				     &CompletionCode,
				     sizeof(CompletionCode));

			DPRINTK("invalid command: FN_CANPOS_READ_CONFIG\n");
			break;
		}
		pBuffer = (unsigned char *) Canpos_read_config_cmd;
		CanposState = CANPOS_COLLECT_READ_CONFIG_DATA;
		CanposConfigBytesCollected = 0;
		canposConfigAck = 0;

		DPRINTK("FN_CANPOS_READ_CONFIG: send config read command\n");

		rc = CanposPosKbdSendData(pBuffer,
					  sizeof(Canpos_read_config_cmd));
		if (rc == 0) {
			DPRINTK("FN_CANPOS_READ_CONFIG: send success\n");
		} else {
			DPRINTK("FN_CANPOS_READ_CONFIG: send failed!!rc=%x\n",
				rc);

			CanposState = CANPOS_READY; 
			readConfigErr = 1;
		}

/*  gha - PROBLEM:  SEV 1 APAR HERE if State never set to READY!!!!!  */
		while (CanposState == CANPOS_COLLECT_READ_CONFIG_DATA) {
			/* wait here */
			if (CanposState == CANPOS_READY)
				break;
		}

		/* Convert and copy canpos configuration data to user data
		 * configuration data start at offset 9 (or 10 for new code) and
		 * is converted to specific format before transmitt.
		 * For example, byte 0xab is converted to 0x9a and 0xab, byte 00
		 * is converted to 0x90 and 0xa0
		 */
		j = canposConfigAck + 6;    /* index to user data */
					    /* 9 for older code (with 3 ack), */
					    /* 10 for code with 4 ack */
		i = 0;
		pCanposReadConfig = (PPOS_CANPOS_READ_CONFIG)pInputBuffer;
		pCanposReadConfig->length = CANPOS_READ_CONFIG_SIZE;

		while (i <  pCanposReadConfig->length) {
			/* even bytes */
			pCanposReadConfig->Data[j] = (canpos_config.Data[i++] & 0x0f) << 4;
			/* odd bytes */
			pCanposReadConfig->Data[j] |= (canpos_config.Data[i++] & 0x0f);
			j++;
		}

		wake_up_canpos_process();
		if (readConfigErr)
			CompletionCode = RPERR;
		else
			CompletionCode = RPDONE;

		copy_to_user(&((POS_CANPOS_READ_CONFIG *) userBuffer)->CompletionCode,
			     &CompletionCode, sizeof(CompletionCode));

		DPRINTK("exit FN_CANPOS_READ_CONFIG\n");

		break;
	}

    case FN_CANPOS_WRITE_CONFIG:
    {
		char Canpos_write_config_cmd[] = {0xE3, 0xC1, 0xC2, 0xC3};

		DPRINTK("FN_CANPOS_WRITE_CONFIG : Enter.\n");

		if (CanposKeyboard != 1) {
			/* invalid Keyboard */
			rc = -EINVAL;
			EPRINTK("FN_CANPOS_WRITE_CONFIG: Not a Canpos KBD.\n");
			break;
		}
		writeConfigErr = 0;

		pCanposWriteConfig = (PPOS_CANPOS_WRITE_CONFIG)pInputBuffer;
		LRC = 0;

		/* let computer the lrc */
		/* XAC simply xor all bytes -- !!!!! */
        for (i = 0; i < W_CONFIG_SIZE; i++)
            LRC = (unsigned char) (LRC ^ pCanposWriteConfig->Data[i]);

        pCanposWriteConfig->Data[0] = (unsigned char) LRC;

		
        //Sending Config Command
        pBuffer = (unsigned char *) Canpos_write_config_cmd;

        CanposState = CANPOS_COLLECT_NORMAL_ACK;

		rc = CanposPosKbdSendData(pBuffer,sizeof(Canpos_write_config_cmd));
		if (rc != 0) {
		    EPRINTK("FN_CANPOS_WRITE_CONFIG: Send config cmd failed!! rc=%x\n", rc);
            CanposState = CANPOS_READY;
            rc = -EINVAL;
		    break;
		}

		if (CanposState != CANPOS_GOT_CONFIG_KEY_1)
        {
            EPRINTK("FN_CANPOS_WRITE_CONFIG: Error-Not Rxed KeyPress 1.\n");
            CanposState=CANPOS_READY;
            rc = -EINVAL;
	    	break;
        }

        //write cofiguration data

        canposConfigAck = 0;
		CanposState = CANPOS_COUNT_WRITE_CONFIG_ACK;

        pCanposWriteConfig->Data[W_CONFIG_SIZE] = 0xE5;

        rc = CanposPosKbdSendData(pCanposWriteConfig->Data, (int)1281);

		if (rc != 0) {
			EPRINTK("FN_CANPOS_WRITE_CONFIG: Send config data Failed.!!rc=%x\n", rc);
            CanposState = CANPOS_READY;
            canposConfigAck = 0;
            rc = -EINVAL;
			break;
		}

		mdelay(1000);

        if (CanposState != CANPOS_GOT_CONFIG_KEY_1)
        {
            EPRINTK("FN_CANPOS_WRITE_CONFIG: Error-Not Rxed KeyPress 1.\n");
            CanposState=CANPOS_READY;
            canposConfigAck = 0;
            rc = -EINVAL;
            break;
        }

        rc = SUCCESS;

        CanposState=CANPOS_READY;
        canposConfigAck = 0;

        DPRINTK("FN_CANPOS_WRITE_CONFIG: Exit.\n");

		break;

	}

	case  FN_CANPOS_FLASH_UPDATE:
	{
		char Canpos_flash_cmd[] = {0xE6, 0xC1, 0xC2, 0xC3};

		DPRINTK("FN_CANPOS_FLASH_UPDATE: Enter.\n");

		if (CanposKeyboard != 1) {
			/* invalid command */
			EPRINTK("FN_CANPOS_FLASH_UPDATE: Not a Canpos KBD.\n");
            rc = -EINVAL;
			break;
		}

		pCanposFlash = (PPOS_CANPOS_FLASH_UPDATE)pInputBuffer;
		flashSize = pCanposFlash->length;

		DPRINTK("FN_CANPOS_FLASH_UPDATE: flash size= %d\n", flashSize);

		// allocate memory to hold flash data + 2 bytes length * 1 byte LRC
		if (!(flashData = (unsigned char *)vmalloc(flashSize+3))) {
			EPRINTK("FN_CANPOS_FLASH_UPDATE: Can't allocate memory for flash");
            rc = -EINVAL;
			break;
		}

        if (copy_from_user(&flashData[2], pCanposFlash->Data,flashSize)) {
			EPRINTK("FN_CANPOS_FLASH_UPDATE:Can't copy user's flash data\n");
			rc = -EINVAL;
            /* free memory */
		    if (flashData != NULL)
			    vfree(flashData);
			break;
		}

        // add size bytes
        flashData[0] = (unsigned char)(flashSize>>8);
		flashData[1] = (unsigned char)flashSize;
		DPRINTK("FN_CANPOS_FLASH_UPDATE: size hi= %x, size lo= %x\n",flashData[0], flashData[1]);
       
		// computer the lrc
		// XAC simply xor all bytes -- !!!!!

		LRC = 0;
        for (i = 0; i < (flashSize+2); i++)
			LRC = (unsigned char) (LRC ^ flashData[i]);
        flashData[flashSize+2] = (unsigned char) LRC;

		// send flash cmd
		pBuffer = (unsigned char *)Canpos_flash_cmd;

        CanposState = CANPOS_COLLECT_NORMAL_ACK;

		rc = CanposPosKbdSendData(pBuffer,sizeof(Canpos_flash_cmd));
		if (rc != 0) {
		    EPRINTK("FN_CANPOS_FLASH_UPDATE: Send flash cmd failed!!rc=%x\n", rc);
            CanposState = CANPOS_READY;
            rc = -EINVAL;
			/* free memory */
		    if (flashData != NULL)
			    vfree(flashData);
            break;
		}

		if (CanposState != CANPOS_GOT_CONFIG_KEY_1)
        {
            EPRINTK("FN_CANPOS_FLASH_UPDATE: Error-Not Rxed KeyPress 1.\n");
            CanposState=CANPOS_READY;
            rc = -EINVAL;
            /* free memory */
	        if (flashData != NULL)
		        vfree(flashData);
		    break;
        }

		// Up to this point, every thing ok, send  flash data
		// keep up with the acks

		canposFlashAck = 0;
		CanposState = CANPOS_COUNT_FLASH_ACK;

		DPRINTK("FN_CANPOS_FLASH_UPDATE: send flash data\n");

		rc = CanposPosKbdSendData(flashData, (int)(flashSize+3));

        if (rc != 0) {
            EPRINTK("FN_CANPOS_FLASH_UPDATE: Send flash data failed!!rc=%x\n", rc);
            CanposState = CANPOS_READY;
            canposFlashAck = 0;
            rc = -EINVAL;
            /* free memory */
		    if (flashData != NULL)
			    vfree(flashData);
		    break;
		}

		mdelay(5000);

		if (CanposState != CANPOS_GOT_CONFIG_KEY_1)
        {
			EPRINTK("FN_CANPOS_FLASH_UPDATE: Error-Not Rxed KeyPress 1.\n");
            canposFlashAck = 0;
            CanposState = CANPOS_READY;
            rc = -EINVAL;
            /* free memory */
		    if (flashData != NULL)
			    vfree(flashData);
			break;

        }

        rc = SUCCESS;

		canposFlashAck = 0;
        CanposState = CANPOS_READY;
        
        /* free memory */
		if (flashData != NULL)
			vfree(flashData);

		DPRINTK("FN_CANPOS_FLASH_UPDATE :Exit.\n");

        break;
	}

	case FN_QUERY_CANPOS_KEYBOARD:
	{
	    unsigned char Canpos_query_config_cmd[] = {0xE1, 0xC5, 0xCA};
        unsigned char Canpos_query_config_cmd1[] = {0xE2};
        int i;

        DPRINTK("FN_QUERY_CANPOS_KEYBOARD: Enter.\n");

        //Send Ver Info Command.
        pBuffer = (unsigned char *)Canpos_query_config_cmd;

        CanposState = CANPOS_COLLECT_NORMAL_ACK;

        rc = CanposPosKbdSendData(pBuffer, 3);

        if (rc != 0) {
            EPRINTK("FN_QUERY_CANPOS_KEYBOARD: Send Ver Info cmd failed!! rc=%x\n", rc);
            CanposState = CANPOS_READY;
            rc = -EINVAL;
	        break;
	    }

	    if (CanposState != CANPOS_GOT_CONFIG_KEY_1)
        {
            EPRINTK("FN_QUERY_CANPOS_KEYBOARD: Error-Not Rxed KeyPress 1.\n");
            CanposState=CANPOS_READY;
            rc = -EINVAL;
			break;
        }


        //Send Ver Info Command 2
        CanposState = CANPOS_COLLECT_VER_INFO_START;
        verinfosize =0;
        pBuffer = (unsigned char *)Canpos_query_config_cmd1;

	    rc = CanposPosKbdSendData(pBuffer, 1);

	    if (rc != 0) {
	        EPRINTK("FN_QUERY_CANPOS_KEYBOARD: Send config data Failed- rc=%x\n", rc);
            CanposState = CANPOS_READY;
            verinfosize =0;
            rc = -EINVAL;
	        break;
	    }

		mdelay(1000);

        if (CanposState != CANPOS_COLLECT_VER_INFO_COLLECTED)
        {
            EPRINTK("FN_QUERY_CANPOS_KEYBOARD: Error-Collecting Version Info.\n");
            verinfosize = 0;
            CanposState = CANPOS_READY;
            rc = -EINVAL;
			break;
        }

        for(i=0;i<VER_INFO_SIZE;i++)
            tmpCanposQueryKeyboard.data[i] = verinfo[i];

        verinfosize = 0;
        CanposState = CANPOS_READY;

        copy_to_user ((MODQUERY_KEYBOARD_PARMS *)pInputBuffer,(unsigned char *) &tmpCanposQueryKeyboard,sizeof(MODQUERY_KEYBOARD_PARMS));

        DPRINTK("FN_QUERY_CANPOS_KEYBOARD: Exit.\n");

        rc = SUCCESS;

        break;

	}

    case FN_POS_LEDS:
	{
		DPRINTK("FN_POS_LEDS\n");

		/* Setup Pos Leds command to send. */
		setIndicators.Command = POS_LEDS_COMMAND;
		setIndicators.Parm1   = ((PPOSLEDS) pInputBuffer)->posLEDs;

		/* Get State of Leds from Linux. */

		/* set up the state of caps_lock, num_lock, scroll_lock. */
		setIndicators.Parm1 |= (unsigned char) (pPosKbdData->KeyboardIndicators.LedFlags);


		/* Save the LEDs state.  I don't think we need to maintain
		 * the state of POS Leds.
		 */

		/* pPosKbdData->KeyboardIndicators.LedFlags =
		 *     ((PPOSLEDS) pInputBuffer->Data)->posLEDs;
		 *    (unsigned char) (pPosKbdData->KeyboardIndicators.LedFlags << 4);
		 */
		pBuffer = (unsigned char *) &setIndicators;

		rc = PosKbdSendData(pBuffer, sizeof(setIndicators));

		CompletionCode = RPDONE;

		copy_to_user(&((POSLEDS *) userBuffer)->CompletionCode,
			     &CompletionCode,
			     sizeof(CompletionCode));

		break;
	}

	case FN_TONE:
	{
		DPRINTK("FN_TONE\n");

		/* Setup Tone command to send. */
		setTone.Command = POS_TONE_COMMAND;
		setTone.Parm1 = ((PTONE) pInputBuffer)->tone1;
		setTone.Parm2 = ((PTONE) pInputBuffer)->tone2;

		/* If Tone ON or Tone OFF (0x60), then send 2 bytes, otherwise
		 * send three bytes.  The third byte contains the tone duration.
		 */
		bytesToSend = (setTone.Parm1 & 0x60) ? 2 : 3;

		pBuffer = (unsigned char *) &setTone;

		rc = PosKbdSendData(pBuffer, bytesToSend);

		CompletionCode = RPDONE;

		copy_to_user(&((TONE *) userBuffer)->CompletionCode,
			     &CompletionCode,
			     sizeof(CompletionCode));

		break;
	}

	case FN_CLICK:
	{
		DPRINTK("FN_CLICK\n");

		switch (((PCLICK) pInputBuffer)->fClick) {
		case PosOFF:
			pPosKbdData->KbdConfigureCmd[1] &= ~(POS_CLICK_LOUD |
							     POS_CLICK_SOFT);
			break;

		case PosSOFT:
			pPosKbdData->KbdConfigureCmd[1] &= ~(POS_CLICK_LOUD);
			pPosKbdData->KbdConfigureCmd[1] |= (POS_CLICK_SOFT);
			break;

		case PosLOUD:
			pPosKbdData->KbdConfigureCmd[1] &= ~(POS_CLICK_SOFT);
			pPosKbdData->KbdConfigureCmd[1] |= (POS_CLICK_LOUD);
			break;
		}

		/* Copy configure command into local buffer. */
		setConfig[0] = pPosKbdData->KbdConfigureCmd[0];
		setConfig[1] = pPosKbdData->KbdConfigureCmd[1];
		setConfig[2] = pPosKbdData->KbdConfigureCmd[2];
		setConfig[3] = pPosKbdData->KbdConfigureCmd[3];

		pBuffer = (unsigned char *) &setConfig;

		rc = PosKbdSendData(pBuffer, sizeof(setConfig));

		CompletionCode = RPDONE;

		copy_to_user(&((CLICK *) userBuffer)->CompletionCode,
			     &CompletionCode,
			     sizeof(CompletionCode));

		break;
	}

	case FN_TYPEMATIC:
	{
		DPRINTK("FN_TYPEMATIC\n");

		/* First save the info into device extension. */
		if (((PTYPEMATIC) pInputBuffer)->fTypematic) {
			pPosKbdData->KbdConfigureCmd[1] |= POS_TYPEMATIC_ON;
		} else {
			pPosKbdData->KbdConfigureCmd[1] &= ~POS_TYPEMATIC_ON;
		}

		/* First copy what is in the configure command. */
		setConfig[0] = pPosKbdData->KbdConfigureCmd[0];
		setConfig[1] = pPosKbdData->KbdConfigureCmd[1];
		setConfig[2] = pPosKbdData->KbdConfigureCmd[2];
		setConfig[3] = pPosKbdData->KbdConfigureCmd[3];

		pBuffer = (unsigned char *) &setConfig;

		rc = PosKbdSendData(pBuffer, sizeof(setConfig));

		CompletionCode = RPDONE;

		copy_to_user(&((TYPEMATIC *) userBuffer)->CompletionCode,
			     &CompletionCode,
			     sizeof(CompletionCode));

		break;
	}

	case FN_ENABLE_DISABLE:
	{
		DPRINTK("FN_ENABLE_DISABLE\n");

		/* First save the info into device extension. */
		if (((PENABLE_DISABLE) pInputBuffer)->fEnable) {
			pPosKbdData->KbdConfigureCmd[1] |= POS_SCANNING_ON;
		} else {
			pPosKbdData->KbdConfigureCmd[1] &= ~POS_SCANNING_ON;
		}

		/* First copy what is in the configure command. */
		setConfig[0] = pPosKbdData->KbdConfigureCmd[0];
		setConfig[1] = pPosKbdData->KbdConfigureCmd[1];
		setConfig[2] = pPosKbdData->KbdConfigureCmd[2];
		setConfig[3] = pPosKbdData->KbdConfigureCmd[3];

		pBuffer = (unsigned char *) &setConfig;

		rc = PosKbdSendData(pBuffer, sizeof(setConfig));

		CompletionCode = RPDONE;

		copy_to_user(&((ENABLE_DISABLE *) userBuffer)->CompletionCode,
			     &CompletionCode,
			     sizeof(CompletionCode));

		break;
	}

	case FN_TRACKS:
	{
		DPRINTK("FN_TRACKS\n");

		setTracks[0] = pPosKbdData->MsrConfigureCmd[0];

		setTracks[1] = (unsigned char) ((PTRACKS) userBuffer)->track1enabled
			       ? 0x1 : 0x00;

		setTracks[1] |= (unsigned char) ((PTRACKS) userBuffer)->track2enabled
				? 0x2 : 0x00;

		setTracks[1] |= (unsigned char) ((PTRACKS) userBuffer)->track3enabled
				? 0x4 : 0x00;

		setTracks[1] |= (unsigned char) ((PTRACKS) userBuffer)->trackJenabled
				? 0x8 : 0x00;

		rc = PosKbdSendData((unsigned char *) &setTracks, sizeof(setTracks));

		CompletionCode = RPDONE;

		copy_to_user(&((TRACKS *) userBuffer)->CompletionCode,
			     &CompletionCode,
			     sizeof(CompletionCode));

		break;
	}

	case FN_TRAP_KEYS:
	{
		DPRINTK("FN_TRAP_KEYS\n");

		pPosKbdData->fTrap = ((PTRAP_KEYS) pInputBuffer)->fTrapKeys;

		rc = SUCCESS;

		CompletionCode = RPDONE;

		copy_to_user(&((TRAP_KEYS *) userBuffer)->CompletionCode,
			     &CompletionCode,
			     sizeof(CompletionCode));

		break;
	}

	case FN_SET_DOUBLEKEYS:
	{	   

        IPRINTK("FN_SET_DOUBLEKEYS\n");
		
        memset(&pPosKbdData->doubleKey, 0,
		       MAX_DOUBLE_KEYS * sizeof(DOUBLE_KEY));

		pPosKbdData->nDoubleKeys = ((PDOUBLE_KEYS) pInputBuffer)->length;

		DPRINTK("Double Keys: total=%d\n", pPosKbdData->nDoubleKeys);

		pParms = (PDOUBLE_KEYS) pInputBuffer;

		i = 0;
		j = 0;

		while (j < pPosKbdData->nDoubleKeys) {
			pPosKbdData->doubleKey[i].primaryKey = pParms->buffer[j++];
			pPosKbdData->doubleKey[i].secondaryKey = pParms->buffer[j++];
			pPosKbdData->doubleKey[i].primaryMake = 0;
			pPosKbdData->doubleKey[i].secondaryMake = 0;
			pPosKbdData->doubleKey[i].lastKey = 0;

			/* Need to set flag if key switches 109 & 108 are doubled
			 */
			if ((pPosKbdData->doubleKey[i].primaryKey == 109) &&
			    (pPosKbdData->doubleKey[i].secondaryKey == 108)) {
				pPosKbdData->fStatus |= DOUBLE_KEYS_109_108;
			} else {
				pPosKbdData->fStatus &= ~DOUBLE_KEYS_109_108;
			}

			DPRINTK("Double Keys - uBuf Key 1 = %d, Key 2 = %d\n",
			       pParms->buffer[j-2], pParms->buffer[j-1]);

			DPRINTK("Double Keys: kBuf Key 1 = %d, Key 2 = %d \n",
			        pPosKbdData->doubleKey[i].primaryKey,
				pPosKbdData->doubleKey[i].secondaryKey);
			i++;
		}

		/* Reset it to actual number of double keys. */
		pPosKbdData->nDoubleKeys = i;

		rc = SUCCESS;

		CompletionCode = RPDONE;

		copy_to_user(&((DOUBLE_KEYS *) userBuffer)->CompletionCode,
			     &CompletionCode,
			     sizeof(CompletionCode));

		break;
	}

	case FN_GET_SHIFT_STATE:
	{
		DPRINTK("FN_GET_SHIFT_STATE\n");

		/* This function implemented in USER MODE, but not
		 * in KERNEL MODE.
		 */
		rc = SUCCESS;
		CompletionCode = RPDONE;

		copy_to_user(&((SHIFT_STATE *) userBuffer)->CompletionCode,
			     &CompletionCode,
			     sizeof(CompletionCode));

		break;
	}

	case FN_SET_SHIFT_STATE:
	{
		DPRINTK("FN_SET_SHIFT_STATE\n");

		/* This function is implemnted in USER MODE  - not in KERNEL.
		 * The driver obtains initial state of LEDs from user mode from
		 * aipanpos utility.
		 */
		pPosKbdData->KeyboardIndicators.LedFlags = (((PSHIFT_STATE) pInputBuffer)->state);

		rc = SUCCESS;
		CompletionCode = RPDONE;

		copy_to_user(&((SHIFT_STATE *) userBuffer)->CompletionCode,
			     &CompletionCode,
			     sizeof(CompletionCode));

		break;
	}

	case FN_QUERY_KEYBOARD:
	{
        #ifdef DEBUG
          printk(KERN_DEBUG "kb_ioctl():FN_QUERY_KEYBOARD:\n");
        #endif

        if (pPosKbdData->PosKbdDeviceInfo.deviceID)
         {
            tmpQueryKeyboard.keyboardID = OB_KBD;

            // Crystal 8x4 keypad support. Crystal DevID = 0x08

            if (pPosKbdData->PosKbdDeviceInfo.deviceID == 0x08)
                tmpQueryKeyboard.keyboardSubtype =  4;    /* crystal keypad */
            else if (pPosKbdData->PosKbdDeviceInfo.deviceID == 0x05)
                tmpQueryKeyboard.keyboardSubtype =  2;    /* ANKPOS  */
         }

        tmpQueryKeyboard.msrType = pPosKbdData->PosMsrDeviceInfo.deviceID;
        tmpQueryKeyboard.keyboardEC = pPosKbdData->PosKbdDeviceInfo.ECLevel;
        tmpQueryKeyboard.msrEC = pPosKbdData->PosMsrDeviceInfo.ECLevel;

        if (CanposKeyboard)
         {
            printk(KERN_DEBUG "kb_ioctl():FN_QUERY_KEYBOARD: Found a CANPOS Keybaord.\n");
            tmpQueryKeyboard.keyboardSubtype = 3;         /* CANPOS */
         }
        else
         {
           switch( pPosKbdData->PosKbdDeviceInfo.Features )
             {
                 case  0x04:

                       //tmpQueryKeyboard.keyboardSubtype = 2;     /* ANKPOS */
                       break;

                 case  0x03:

                       tmpQueryKeyboard.keyboardSubtype =  1;    /* NANPOS */
                       break;

             }
         }

        copy_to_user ( pInputBuffer,
                       (unsigned char *) &tmpQueryKeyboard,
                       sizeof(QUERY_KEYBOARD_PARMS));

        rc = SUCCESS;

        CompletionCode = RPDONE;

        copy_to_user(&((QUERY_KEYBOARD_PARMS *) userBuffer)->CompletionCode,
                     &CompletionCode,
                     sizeof(CompletionCode));

        break;
	}

	case FN_READ_KBD_STATUS:
	{
		/* copy STATUS data */
		copy_to_user(((READ_KBD_STATUS_PARMS *)userBuffer)->buffer,
			      &pPosKbdData->PosKbdStatus.KBDStatus, 4);

		rc = SUCCESS;
		CompletionCode = RPDONE;

		copy_to_user(&((QUERY_KEYBOARD_PARMS *) userBuffer)->CompletionCode,
			     &CompletionCode,
			     sizeof(CompletionCode));

		break;
	}

	case FN_READ_MSR_DATA:
	{
		/* DPRINTK("FN_READ_MSR_DATA\n"); */

		if (pPosKbdData->MSRData.MSRDataLength &&
		    !(pPosKbdData->fStatus & COLLECTING_MSR_STATUS)) {

			DPRINTK("reading MSR data %ld bytes\n",
				pPosKbdData->MSRData.MSRDataLength);

			copy_to_user(((PREAD_MSR_DATA_PARMS)pInputBuffer)->buffer,
				     &pPosKbdData->MSRData.MSRDataBuffer,
				     pPosKbdData->MSRData.MSRDataLength);

			copy_to_user(&(((PREAD_MSR_DATA_PARMS) pInputBuffer)->length),
				     &pPosKbdData->MSRData.MSRDataLength,
				     sizeof(pPosKbdData->MSRData.MSRDataLength));

			DPRINTK("MSR Buffer 0x%x, 0x%x, 0x%x, 0x%x, 0x%x, 0x%x, 0x%x, ...\n",
			       pPosKbdData->MSRData.MSRDataBuffer[0],
			       pPosKbdData->MSRData.MSRDataBuffer[1],
			       pPosKbdData->MSRData.MSRDataBuffer[2],
			       pPosKbdData->MSRData.MSRDataBuffer[3],
			       pPosKbdData->MSRData.MSRDataBuffer[4],
			       pPosKbdData->MSRData.MSRDataBuffer[5],
			       pPosKbdData->MSRData.MSRDataBuffer[6]);

			/* Reset MSR Data Length */
			memset(&pPosKbdData->MSRData, 0,
			       sizeof(pPosKbdData->MSRData));

			pPosKbdData->MSRData.MSRDataLength = 0;

			DPRINTK("FN_READ_MSR_DATA\n");
		} else {
			copy_to_user(&((READ_MSR_DATA_PARMS *) userBuffer)->length,
				     &Length,
				     sizeof(Length));
		}

		rc = SUCCESS;
		CompletionCode = RPDONE;

		copy_to_user(&((READ_MSR_DATA_PARMS *) userBuffer)->CompletionCode,
			     &CompletionCode,
			     sizeof(CompletionCode));

		break;
	}

	case FN_CODE_UPDATE:
	{
		DPRINTK("FN_CODE_UPDATE\n");

		/* Codeupdate format of the data received is:
		 *
		 *   o | length | data | length | data | length | data |
		 *   o The length of zero indicates end of code update.
		 *
		 * Data format sent to the keyboard must be:
		 *
		 *   o | 0x22 (cmd) | length | data |
		 *   o | 0x22 (cmd) | 0x00 | must be sent to end code update.
		 */

		/*
		 * Initilize Buffer: Buffer[0] = 0x22 (update command)
		 *                   Rest of buffer to zero
		 */
		pPosKbdData->fCodeUpdtDone          = 0;
		pPosKbdData->CodeUpdate.iUpdate     = 0;
		pPosKbdData->CodeUpdate.iBuffer     = 0;
		pPosKbdData->CodeUpdate.State       = 0;
		pPosKbdData->CodeUpdate.EndOfaBlock = 0;

		for (i = 0;  i < UPDATE_SIZE; i++) {
			pPosKbdData->CodeUpdate.Update[i].Buffer[0] = 0x22;
			pPosKbdData->CodeUpdate.Update[i].Length    = 2;

			for (j = 1; j < 129; j++) {
				pPosKbdData->CodeUpdate.Update[i].Buffer[j] = 0;
			}
		}

		/* Copy the data into devince extension in the correct format.
		 */
		pCurUpdate =  (unsigned char *) ((PCODE_UPDATE_DATA)pInputBuffer)->Data;
		CurLength  = *pCurUpdate;

		if (CurLength != 0) {

			i = 0;
			pNextUpdate = pCurUpdate + (CurLength + 1);

			while (CurLength != 0) {
			       pPosKbdData->CodeUpdate.Update[i].Length = CurLength;

				/* Need to copy (Length + 1) bytes to include
				 * length byte.
				 */
				memcpy(&pPosKbdData->CodeUpdate.Update[i].Buffer[1],
					pCurUpdate,
					CurLength+1);

				DPRINTK("Block=0x%x, Data: [0]=0x%x,[1]=0x%x, [2]=0x%x, [N-1]=0x%x, [N]=0x%x\n",
				       i,
				       pPosKbdData->CodeUpdate.Update[i].Buffer[0],
				       pPosKbdData->CodeUpdate.Update[i].Buffer[1],
				       pPosKbdData->CodeUpdate.Update[i].Buffer[2],
				       pPosKbdData->CodeUpdate.Update[i].Buffer[CurLength],
				       pPosKbdData->CodeUpdate.Update[i].Buffer[CurLength+1]);

				i++;
				pCurUpdate = pNextUpdate;
				pNextUpdate += *pNextUpdate + 1;
				CurLength = *pCurUpdate;
			}
		}

		/* Now insert an ioctl to handle the actual code update. */
		DPRINTK("Send Code Update.\n");

		/* Process the code update. */
		pos_process_code_update();

		rc = SUCCESS;
		CompletionCode = RPDONE;

		copy_to_user(&((CODE_UPDATE_DATA *) userBuffer)->CompletionCode,
			     &CompletionCode,
			     sizeof(CompletionCode));

		break;
	}

	case FN_QUERY_CODE_UPDATE:
	{
		DPRINTK("FN_QUERY_CODE_UPDATE\n");

		tmpQueryUpdate.fStatus = pPosKbdData->fCodeUpdtDone;

		/* Send the return code through buffer */
		copy_to_user((CODE_UPDATE_STR *)userBuffer, &tmpQueryUpdate,
			     sizeof(CODE_UPDATE_STR));


		DPRINTK("FN_QUERY_CODE_UPDATE  fCodeUpdtDone = 0x%x\n",
		       pPosKbdData->fCodeUpdtDone);

		DPRINTK("FN_QUERY_CODE_UPDATE  status = 0x%x\n",
		       ((PCODE_UPDATE_STR) pInputBuffer)->fStatus);

		rc = SUCCESS;

		CompletionCode = RPDONE;

		copy_to_user(&((CODE_UPDATE_STR *) userBuffer)->CompletionCode,
			     &CompletionCode,
			     sizeof(CompletionCode));

		break;
	}

	case ENABLE_ANPOS_ASYNC_COMM:
	{
		ENABLE_ANPOS_ASYNC_COMM_PARMS Parms;

		DPRINTK("ENABLE_ANPOS_ASYNC_COMM\n");

		rc = SUCCESS;

		/* Save process ID to notify, and set flag. */
		if (copy_from_user(&Parms, (void*)userBuffer, sizeof Parms)) {
			return -EFAULT;
		}

		asyncPID = (int)Parms.ProcessID;
		asyncEnabled = 1;

		Parms.CompletionCode = RPDONE;

		if (copy_to_user((void*)userBuffer, &Parms, sizeof Parms))
			rc = -EFAULT;

		/* Send signal right away, so aipctrl can process any
		 * pending messages.
		 */
		if (asyncEnabled) {
			kill_proc(asyncPID, SIGUSR2, 1);
		}

		break;
	}

	case DISABLE_ANPOS_ASYNC_COMM:
	{
		DPRINTK("DISABLE_ANPOS_ASYNC_COMM\n");

		rc = SUCCESS;

		/* reset flag and processID. */
		asyncPID = 0;
		asyncEnabled = 0;

		completionCodeParms.CompletionCode = RPDONE;

		if (copy_to_user((ANPOS_COMPLETION_CODE_PARMS *) userBuffer,
				 &completionCodeParms,
				 sizeof(ANPOS_COMPLETION_CODE_PARMS)))
			rc = -EFAULT;

		break;
	}

	default:
	{
		DPRINTK("Invalid IOCTL, CMD=%d\n", cmd);
		rc = -EINVAL;
	}
	}

	return rc;
}


/* :H1.PosKbdSendData() - Send data to keyborad.
 * ***********************************************************************
 *
 * Function Name:  PosKbdSendData()
 *
 * Purpose:        To send data to keyboard.
 *
 * Description:    A common routine to send data to keyboard.
 *
 * Input:          unsigned char * buffer.
 *                 size of buffer
 *
 * Output:         returns number of bytes transferred.
 *                 0 - success
 *                 1 - failure.
 *
 * Notes:
 *
 * ***********************************************************************
 */

static int PosKbdSendData(unsigned char *buf, unsigned char size)
{
	int rc = 0;
	int cmd_retries = 100;  
	int xferred = 0, xfer_retry = size;
	unsigned char b = 0;

	if (size > 1)
		b = *(buf + 1);

	DPRINTK("enter\n");

	while (cmd_busy && cmd_retries) {
		DPRINTK("going to sleep\n");

		set_current_state(TASK_INTERRUPTIBLE);
		schedule_timeout(5);
		--cmd_retries;

		DPRINTK("return from sleep\n");
	}

	/* Call pc_key driver's send routine to send data. */
	cmd_busy = 1;

	do {
		DPRINTK("sending %d bytes to kbd\n", size);

		/* size: number of retires and size of buffer */
		rc = pos_filter.write_data(buf, size, &xferred, size,1000);  

		if (rc != 0) {
			DPRINTK("pos_filter.write_data returned error %d\n", rc);

			/* the kernel patch misses ACK, and does
			 * not increment the xferred count
			 */
			if (rc == (-EINTR))
				xferred++;

			if (xferred < size) {
				buf += xferred;
				size -= xferred;
				xferred = 0;
			}
		}
	} while (xferred < size && xfer_retry--);

	/* Go to sleep.  The wait queue is awakened in pos_kbd_filter() routine
	 * when command is complete i.e. receipt of
	 *
	 *   -  0xF2 end of keyboard status OR
	 *   -  0xF6 end of msr status      OR
	 *   -  0xF7 ROL in response to keyboard reset.
	 */
	if (rc != 0)
		cmd_busy = 0;

	if (cmd_busy != 0)
		interruptible_sleep_on_timeout(&wq, 100);

	DPRINTK("exit, rc=0x%x %s\n", rc, rc == 0 ? "Success" : "Failed");

	return rc;
}


/* :H1.PosKbdReadId() - Read Keyboard ID.
 * ***********************************************************************
 *
 * Function Name:  PosKbdReadId()
 *
 * Purpose:        To read Keyboard ID.
 *
 * Description:    This routine sends Read Id command. The keyboard
 *                 response contains only two bytes of data with no
 *                 framing bytes as it would send for status.
 *                  - Id byte 1 = 0xAB
 *                  - Id byte 2 = 0x87
 *
 * Input:          none
 *
 * Output:         none
 *
 * Notes:
 *
 * ***********************************************************************
 */

static int PosKbdReadId(void)
{
	int rc;
	unsigned char inputBuffer[1];

	DPRINTK("enter\n");

	/* Set up Device Info Command */
	inputBuffer[0] = 0xF2;

	/* Write to keyboard */
	fStatus |= COLLECTING_PRIMARY_ID;

	rc = PosKbdSendData(inputBuffer, sizeof(inputBuffer));

	DPRINTK("exit, rc=0x%x %s\n", rc, rc == 0 ? "Success" : "Failed");

	return rc;

}


/* :H1.PosKbdReqDevInfo() - ReqDevInfo Keyboard.
 * ***********************************************************************
 *
 * Function Name:  PosKbdReqDevInfo()
 *
 * Purpose:        To request device information.
 *
 * Description:    This routine sends device info request to POS keyboard,
 *                 and response is saved.  The routine is called once
 *                 during initialization.
 *
 * Input:          none
 *
 * Output:         none
 *
 * Notes:
 *
 * ***********************************************************************
 */

static int PosKbdReqDevInfo(void)
{
	int rc;
	DEV_INFO_REQ devInfoReq;
	unsigned char *inputBuffer;

	inputBuffer = (unsigned char *) &devInfoReq;

	DPRINTK("enter\n");

	/* Set up Device Info Command */
	devInfoReq.cmd1 = POS_KBD_DEV_INFO_CMD_BYTE1;
	devInfoReq.cmd2 = POS_KBD_DEV_INFO_CMD_BYTE2;
	devInfoReq.cmd3 = POS_KBD_DEV_INFO_CMD_BYTE3;

	/* Write to keyboard */
	rc = PosKbdSendData(inputBuffer, sizeof(devInfoReq));

	DPRINTK("exit, rc=0x%x %s\n", rc, rc == 0 ? "Success" : "Failed");

	return rc;
}


/* :H1.PosMsrReqDevInfo() - ReqDevInfo Keyboard.
 * ***********************************************************************
 *
 * Function Name:  PosMsrReqDevInfo()
 *
 * Purpose:        To request device information.
 *
 * Description:    This routine sends device info request to POS keyboard,
 *                 and response is saved.  The routine is called once
 *                 during initialization.
 *
 * Input:          none
 *
 * Output:         none
 *
 * Notes:
 *
 * ***********************************************************************
 */

static int PosMsrReqDevInfo(void)
{
	int rc;
	DEV_INFO_REQ devInfoReq;
	unsigned char * inputBuffer;

	inputBuffer = (unsigned char *) &devInfoReq;

	DPRINTK("enter\n");

	/* Set up Device Info Command */
	devInfoReq.cmd1 = POS_MSR_DEV_INFO_CMD_BYTE1;
	devInfoReq.cmd2 = POS_MSR_DEV_INFO_CMD_BYTE2;
	devInfoReq.cmd3 = POS_MSR_DEV_INFO_CMD_BYTE3;

	/* Write to keyboard */
	rc = PosKbdSendData(inputBuffer, sizeof(devInfoReq));

	DPRINTK("exit, rc=0x%x %s\n", rc, rc == 0 ? "Success" : "Failed");

	return rc;
}


/* :H1.set_leds-default_state() - set leds to default state.
 * ***********************************************************************
 *
 * Function Name:  set_leds-default_state()
 *
 * Purpose:        To set leds default state.
 *
 * Description:    This routine turns off all leds by default.
 *                 -  pos specific leds.
 *                 -  CapsLock/NumLock/ScrollLock leds also..
 *
 * Input:          none
 *
 * Output:         none
 *
 * Notes:
 *
 * ***********************************************************************
 */

static void set_leds_default_state(void)
{
	int rc;
	KBD_SET_INDICATORS   setIndicators;
	unsigned char       *pBuffer;

	DPRINTK("enter\n");

	/* Setup Pos Leds command to send. */
	setIndicators.Command = POS_LEDS_COMMAND;
	setIndicators.Parm1 = 0;

	/* Get State of Leds from Linux. */

	/* ***********    add   ************ */

	/* set up the state of caps_lock, num_lock, scroll_lock. */
	setIndicators.Parm1 |= (unsigned char) (pPosKbdData->KeyboardIndicators.LedFlags);

	pBuffer = (unsigned char *) &setIndicators;

	rc = PosKbdSendData(pBuffer, sizeof(setIndicators));

	DPRINTK("exit, rc=0x%x %s\n", rc, rc == 0 ? "Success" : "Failed");
}


/* :H1.SetDefaultConfiguration() - Send default configuration..
 * ***********************************************************************
 *
 * Function Name:  SetDefaultConfiguration()
 *
 * Purpose:        To send default configuration.
 *
 * Description:    This routine sends default configuration to keyboard.  
 *                 By default, it turns on typematic. 
 *
 * Input:          none
 *
 * Output:         rc = 0 success
 *
 * Notes:
 *
 * ***********************************************************************
 */

static int SetDefaultConfiguration(void)
{
	int rc;
	unsigned char setConfig[KBD_CONFIGURE_CMD_LENGTH];

	DPRINTK("enter\n");

	/* Copy configure command into local buffer. */
	setConfig[0] = pPosKbdData->KbdConfigureCmd[0];
	setConfig[1] = pPosKbdData->KbdConfigureCmd[1];
	setConfig[2] = pPosKbdData->KbdConfigureCmd[2];
	setConfig[3] = pPosKbdData->KbdConfigureCmd[3];

	/* Write to keyboard */
	rc = PosKbdSendData(setConfig, sizeof(setConfig));

	DPRINTK("exit, rc=0x%x %s\n", rc, rc == 0 ? "Success" : "Failed");

	return rc;
}


#if defined(KBD_RESET)
/* :H1.PosKbdReset() - Reset Keyboard.
 * ***********************************************************************
 *
 * Function Name:  PosKbdReset()
 *
 * Purpose:        To reset the keyboard.
 *
 * Description:    This routine sends reset to POS keyboard, and is called
 *                 during driver initialization.
 *
 * Input:          none
 *
 * Output:         none
 *
 * Notes:
 *
 * ***********************************************************************
 */

static int PosKbdReset(void)
{
	int rc;
	unsigned char inputBuffer[2] = {0, 0};

	DPRINTK("enter\n");

	/* Set up Device Info Command */
	inputBuffer[0] = (unsigned char) POS_KBD_RESET_CMD_BYTE1;
	inputBuffer[1] = (unsigned char) POS_KBD_RESET_CMD_BYTE2;

	/* Write to keyboard */
	rc = PosKbdSendData(inputBuffer, sizeof(inputBuffer));

	DPRINTK("exit, rc=0x%x %s\n", rc, rc == 0 ? "Success" : "Failed");

	return rc;
}
#endif


/* :H1.pos_process_code_update - Processes CodeUpdate.
 * ********************************************************************
 *
 * Function Name:  pos_process_code_update()
 *
 * Purpose:        Supports CodeUpdate for POS Keybaord.
 *
 * Description:    This is routine processes code updates synchronously.
 *                 It builds an irp for each code update block and sends
 *                 it down the stack.  The routines exits when it senses
 *                 the code update complete bit.
 *
 * Input:          Target - The top of the stack
 *
 *                 Ioctl  - The IOCTL to send
 *
 * Output:         status - NTSTATUS
 *
 * Notes:
 *
 * ********************************************************************
 */

static void pos_process_code_update(void)
{
	int rc = 0;

	DPRINTK("enter\n");

	/* Set flag to indicate code update in progress. */
	pPosKbdData->CodeUpdate.State |= SENDING_CODE_UPDATE;
	pPosKbdData->CodeUpdate.State |= CODE_UPDATE_STORED;
	pPosKbdData->CodeUpdate.State &= ~CODE_UPDATE_LAST_BLOCK_SENT;

	while ((pPosKbdData->CodeUpdate.State & SENDING_CODE_UPDATE)) {
		if (pPosKbdData->CodeUpdate.State & CODE_UPDATE_STORED) {
			unsigned int updateBufferLength;

			pPosKbdData->CodeUpdate.State &= ~CODE_UPDATE_STORED;

			memset(updateBuffer, 0, MAX_BUFFER_SIZE);

			/* Setup a block of code update to send. */
			updateBufferLength = pPosKbdData->CodeUpdate.Update[pPosKbdData->CodeUpdate.iUpdate].Length;

			if (updateBufferLength == 2) {
				pPosKbdData->CodeUpdate.State &= ~SENDING_CODE_UPDATE;
				pPosKbdData->CodeUpdate.State |= CODE_UPDATE_LAST_BLOCK_SENT;
			} else {
				updateBufferLength += 2;
			}

			memcpy(updateBuffer,
				&pPosKbdData->CodeUpdate.Update[pPosKbdData->CodeUpdate.iUpdate].Buffer[0],
				updateBufferLength);

			DPRINTK("Sending CodeUpdate Block=0x%x, Length=0x%x\n",
			       pPosKbdData->CodeUpdate.iUpdate,
			       updateBufferLength);
			DPRINTK("Data: 0x%x.0x%x.0x%x.0x%x...0x%x,0x%x\n",
			       updateBuffer[0],
			       updateBuffer[1],
			       updateBuffer[2],
			       updateBuffer[3],
			       updateBuffer[updateBufferLength - 1],
			       updateBuffer[updateBufferLength]);

			/* Write to keyboard */
			rc = PosKbdSendData(updateBuffer, updateBufferLength);

			/* Increment the index to point to next block. */
			pPosKbdData->CodeUpdate.iUpdate++;
		}
	}

	DPRINTK("exit\n");
}


/* :H1.CollectingKBDId() - Collect Keyboard Id
 * ***********************************************************************
 *
 * Function Name:  CollectingKBDId()
 *
 * Purpose:        Save KBD ID bytes
 *
 * Description:    This function saves the KBD Id bytes.
 *                 The keyboard sends only two bytes of id bytes.
 *                   -  ID byte 1 = 0xAB
 *                   -  ID byte 2 = 0x87 
 *
 * Input:          scancode-  the scancode received from the keyboard
 *
 * Output:         scancode - the scancode
 *                  0x00    - if we filter it off
 *
 * Notes:
 *
 * ***********************************************************************
 */

static unsigned char CollectingKBDId(unsigned char scancode)
{
	DPRINTK("scancode=0x%x\n", scancode);

	if (pPosKbdData->PosKbdId.Index < 2) {
		pPosKbdData->PosKbdId.KBDId[pPosKbdData->PosKbdId.Index] = scancode;

		if (pPosKbdData->PosKbdId.Index == 1) {
			fStatus &= ~COLLECTING_PRIMARY_ID;

			DPRINTK("KbdId[1]=0x%x, KbdId[2]=0x%x\n",
			       pPosKbdData->PosKbdId.KBDId[0],
			       pPosKbdData->PosKbdId.KBDId[1]);

			/* wake up the process */
			cmd_busy = 0;
			wake_up_interruptible(&wq);

			DPRINTK("cmd_busy=0x%x \n", cmd_busy);
		} else {
			++pPosKbdData->PosKbdId.Index;
		}
		scancode = 0;
	}

	DPRINTK("returning scancode=0x%x\n", scancode);

	return scancode;
}


/* :H1.CollectingKBDStatus() - Collect Keyboard Status
 * ***********************************************************************
 *
 * Function Name:  CollectingKBDStatus()
 *
 * Purpose:        Save KBD status scancodes
 *
 * Description:    This function saves the KBD status scancodes in the
 *                 KBDStatus buffer.  This function checks to see if there
 *                 is enough room to store the value.  If the value is
 *                 saved, the scancode is set to 0 so that no further
 *                 processing will be done.  The 0xF2 at the end of the
 *                 status message is not stored, but is passed back to
 *                 POSFilter for further processing.
 *
 * Input:          scancode        -  the scancode received from the keyboard
 *
 * Output:         scancode - the scancode
 *                  0x00    - if we filter it off
 *
 * Notes:
 *
 * ***********************************************************************
 */

static unsigned char CollectingKBDStatus(unsigned char scancode)
{
	DPRINTK("scancode=0x%x\n", scancode);

	if (SCANCODE_END_KBD_STATUS != scancode) {
		if (pPosKbdData->PosKbdStatus.Index < 9) {
			pPosKbdData->PosKbdStatus.KBDStatus[pPosKbdData->PosKbdStatus.Index] = convert[scancode];

			if ((OB_KBD == KbdHWIDs) &&
			    pPosKbdData->PosKbdStatus.Index == 2) {
				keyLockState = (unsigned char)((scancode >> 4) & 0x03);
			} else if ((ANPOS_KBD == KbdHWIDs) &&
				   pPosKbdData->PosKbdStatus.Index == 0) {
				keyLockState = (unsigned char)((scancode >> 6) & 0x01);
			}

			++pPosKbdData->PosKbdStatus.Index;
		} else {
			fStatus &= ~(COLLECTING_KBD_STATUS);               
			pPosKbdData->fStatus &= ~COLLECTING_KBD_STATUS;      
		}

		scancode = 0;
	}

	DPRINTK("returning scancode=0x%x\n", scancode);

	return scancode;
}


/* :H1.CollectingMSRStatus()  - Collect MSR Status
 * ********************************************************************
 *
 * Function Name: CollectingMSRStatus
 *
 * Purpose:       Save MSR data (status) scancodes
 *
 * Description:   This function saves the MSR data scancodes in the
 *                MSRData buffer.  This function checks to see if there
 *                is enough room to store the value.  If the value is
 *                saved, the scancode is set to 0 so that no further
 *                processing will be done.  The 0xF6 at the end of the
 *                status message is not stored, but is passed back to
 *                POSFilter for further processing.
 *
 * Input:         scancode - the scancode received from the keyboard
 *
 * Output:        scancode - the scancode
 *                0x00     - if we filter it off
 *
 * Notes:
 *
 * ********************************************************************
 */

static unsigned char CollectingMSRStatus(unsigned char scancode)
{
	DPRINTK("scancode=0x%x\n", scancode);

	if ((SCANCODE_END_MSR_STATUS != scancode) &&
	    (scancode != SCANCODE_ACK)) {

		if (pPosKbdData->MSRData.Index < sizeof(pPosKbdData->MSRData)) {
			/* ***************************************************
			 * Check to see if the previous scancode was a 0xFF.
			 * If so, we need to handle this scancode differently:
			 * ***************************************************
			 */
			if (0xFF == pPosKbdData->MSRData.MSRDataBuffer[pPosKbdData->MSRData.Index]) {
				switch (scancode) {
				case 0xF3:
				case 0xF4:
				case 0xEA:
					pPosKbdData->MSRData.MSRDataBuffer[pPosKbdData->MSRData.Index] = scancode & (unsigned char) 0x7F;

					break;

				default:
					pPosKbdData->MSRData.MSRDataBuffer[pPosKbdData->MSRData.Index] = convert[scancode];

					break;
				}
			} else {
				pPosKbdData->MSRData.MSRDataBuffer[pPosKbdData->MSRData.Index] = convert[scancode];
			}

			/* ***************************************************
			 * If this is a 0xFF, we will not increment the pointer.
			 * The next time through, we will check for 0xFF to see
			 * if we need to process the incoming scancode
			 * differently.  The following is the list of 0xFF
			 * prefixed scancodes:
			 *  0xF3 -> 0x73
			 *  0xF4 -> 0x74
			 *  0xEA -> 0x6A
			 * ***************************************************
			 */
			if (0xFF != scancode) {
				pPosKbdData->MSRData.Index += 1;
				pPosKbdData->MSRData.MSRDataLength += 1;
			}
		} else {
			fStatus &= ~COLLECTING_MSR_STATUS;
			pPosKbdData->fStatus &= ~COLLECTING_MSR_STATUS;
		} 

		scancode = 0;
	}

	DPRINTK("returning, scancode=0x%x\n", scancode);

	return scancode;
}


/* :H1.DoubleKeyProcessing() - Process Double Keys
 * ***********************************************************************
 *
 * Function Name:  DoubleKeyProcessing
 *
 * Purpose:        Handle doublekeys
 *
 * Description:    This function examines the scancode passed in to
 *                 determine if it is part of a doublekey pair.  If it
 *                 is the primary scancode of the doublekey pair, it is
 *                 passed along and a flag is set.  If it is the secondary
 *                 value of the doublekey pair, the primary value is
 *                 passed in and a flag is set.  In either case, if the
 *                 flag has been set already, the scancode is set to 0.
 *
 * Input:          key             - key switch number.
 *                 fBreak          - break code flag
 *                 deviceExtension - Pointer to DevExt.
 *
 * Output:         scancode - the scancode its double pair
 *                 0x00
 *
 * Notes: NONE
 *
 * ***********************************************************************
 */

static unsigned char DoubleKeyProcessing(unsigned char key, unsigned char fBreak)
{
	unsigned short i;

	DPRINTK("double keys: nDoubleKeys 0x%x, key = 0x%x\n",
		pPosKbdData->nDoubleKeys, key);

	for (i = 0; i < pPosKbdData->nDoubleKeys; ++i) {
		if (key == pPosKbdData->doubleKey[i].primaryKey) {
			if (fBreak) {
				pPosKbdData->doubleKey[i].primaryMake = FALSE;

				if (pPosKbdData->doubleKey[i].secondaryMake) {
					pPosKbdData->doubleKey[i].lastKey = key;
					key = 0;
				} else {
					pPosKbdData->doubleKey[ i ].lastKey = 0;
				}
			} else {
				pPosKbdData->doubleKey[i].primaryMake = TRUE;

				if (pPosKbdData->doubleKey[i].lastKey) {
					if (pPosKbdData->doubleKey[i].lastKey != key) {
						pPosKbdData->doubleKey[i].lastKey = key;
						key = 0;
					}
				} else {
					pPosKbdData->doubleKey[i].lastKey = key;
				}
			}
			break;
		} else if (key == pPosKbdData->doubleKey[i].secondaryKey) {
			if (fBreak) {
				pPosKbdData->doubleKey[i].secondaryMake = FALSE;

				if (pPosKbdData->doubleKey[i].primaryMake) {
					pPosKbdData->doubleKey[i].lastKey = key;
					key = 0;
				} else {
					key = pPosKbdData->doubleKey[i].primaryKey;
					pPosKbdData->doubleKey[i].lastKey = 0;
				}
			} else {
				pPosKbdData->doubleKey[ i ].secondaryMake = TRUE;

				if (pPosKbdData->doubleKey[i].lastKey) {
					if (pPosKbdData->doubleKey[i].lastKey != key) {
						pPosKbdData->doubleKey[i].lastKey = key;
						key = 0;
					} else {
						key = pPosKbdData->doubleKey[i].primaryKey;
					}
				} else {
					pPosKbdData->doubleKey[i].lastKey = key;
					key = pPosKbdData->doubleKey[i].primaryKey;
				}
			}
			break;
		}
	}

	DPRINTK("return: nDoubleKeys 0x%x, key=0x%x\n",
		pPosKbdData->nDoubleKeys, key);

	return key;
}


/* :H1.SetShiftKeyState() - Maintain Shift Key States
 * ***********************************************************************
 *
 * Function Name:  SetShiftKeyState()
 *
 * Purpose:        Support the shift states of the keyboard.
 *
 * Description:    This routine saves the shift status of the keys such
 *                 as SHIFT, ALT, CTRL keys.
 *
 * Input:           unsigned char        scanCode,
 *                  POS_KEYBOARD_SCAN_STATE *scanState
 *
 * Output:          none
 *
 * Notes:
 *
 * ***********************************************************************
 */

static void SetShiftKeyState(unsigned char scanCode,
			     POS_KEYBOARD_SCAN_STATE *scanState)
{
	if (scanCode <= (unsigned char) 0x7F) {
		/* make code */
		switch (scanCode) {
		case 0x1D:				      /* 'CTRL' */
			if (*scanState == OK)		      /*  Left */
				pPosKbdData->ShiftFlags |= 0x20;
			else if (*scanState == gotE0)	      /*  Right */
				pPosKbdData->ShiftFlags |= 0x02;
			break;
		case 0x38:				      /* 'ALT' */
			if (*scanState == OK)		      /*  Left */
				pPosKbdData->ShiftFlags |= 0x40;
			else if (*scanState == gotE0)	      /*  Right */
				pPosKbdData->ShiftFlags |= 0x04;
			break;
		case 0x36:				      /* Right 'Shift' */
			if (*scanState == OK)
				pPosKbdData->ShiftFlags |= 0x01;
			break;
		case 0x2A:				      /* Left 'Shift' */
			if (*scanState == OK)
				pPosKbdData->ShiftFlags |= 0x10;
			break;
		default:
			break;
		}
	} else {
		/* break code */
		switch (scanCode & 0x7F) {
		case 0x1D:				      /* 'CTRL' */
			if (*scanState == OK)		      /*  Left */
				pPosKbdData->ShiftFlags &= ~0x320;
			else if (*scanState == gotE0)	      /*  Right */
				pPosKbdData->ShiftFlags &= ~0x302;
			break;
		case 0x38:				      /* 'ALT' */
			if (*scanState == OK)		      /*  Left */
				pPosKbdData->ShiftFlags &= ~0x340;
			else if (*scanState == gotE0)	      /*  Right */
				pPosKbdData->ShiftFlags &= ~0x304;
			break;
		case 0x36:				      /* Right 'Shift' */
			if (*scanState == OK)
				pPosKbdData->ShiftFlags &= ~0x301;
			break;
		case 0x2A:				      /* Left 'Shift' */
			if (*scanState == OK)
				pPosKbdData->ShiftFlags &= ~0x310;
			break;
		default:
			break;
		}
	}

	/* DPRINTK("state = 0x%x\n", pPosKbdData->ShiftFlags); */
}


/* :H1.pos_scancode_filter() - Main Proc To Filter Out POS Keys.
 * ********************************************************************
 *
 * Function Name:  pos_scancode_filter
 *
 * Purpose:        Filter out the POS specific data
 *
 * Description:    This function looks for POS specific data (like MSR
 *                 and status) and filters it before the normal keyboard
 *                 interrupt routine gets a chance to process it.
 *
 * Input:          none
 *
 * Output:
 *
 * Notes:
 *
 * ********************************************************************
 */

static int pos_scancode_filter(unsigned char *pScancode)
{
	unsigned char key = 0;
	unsigned char fBreak = 0xFF;
	int rc = 0;
	POS_KEYBOARD_SCAN_STATE  *scanState;
	unsigned char scancode = *pScancode;
	int devInfoInd;
	unsigned char i;

	DPRINTK("enter scancode=0x%x, *pScan=0x%x\n", scancode, *pScancode);

	scanFlag = 0;
	scanState = &pPosKbdData->CurrentScanState;

	/* *******************************************************************
	 * If read keyboard id is sent, we need to collect next two bytes
	 * from keyborad.  Start collecting after we receive ACK (0xFA).
	 * Note that keyboard does not send any framing bytes for readID 
	 * command unlike it would for other Pos Keyboards commands.
	 * Checking for 0x9c (break code for enter key) also it represents part 
	 * of the enter key that was hit.  I have see this break code come in
	 * during cmd/response processing.
	 * *******************************************************************
	 */
	if ((fStatus & COLLECTING_PRIMARY_ID) &&
	    (scancode != ACKNOWLEDGE) && (scancode != 0x9c)) {
		scancode = CollectingKBDId(scancode);
		scanFlag = 1;
	}

	/* *******************************************************************
	 * If we have received a 0xF5, save the incoming scancodes in the MSR
	 * data buffer until we receive a 0xF6.  When we receive the 0xF6, we
	 * will send a notification scancode through (see the switch below).
	 * *******************************************************************
	 */
	if ((scancode != SCANCODE_ACK) &&
	    (scancode != SCANCODE_RESEND) && (scancode != SCANCODE_ROL)) {
		if (fStatus & COLLECTING_MSR_STATUS) {
			scancode = CollectingMSRStatus(scancode);
			scanFlag = 1;
		}

		/* ***********************************************************
		 * If we have received a 0xF1 and have not yet received a 0xF2,
		 * we collect the status bytes in the buffer.
		 * ***********************************************************
		 */
		else if (scancode && (fStatus & COLLECTING_KBD_STATUS)) {
			scancode = CollectingKBDStatus(scancode);
			scanFlag = 1;
		}
	}

    if((CanposKeyboard == 1) && (scanFlag==0))
    {
        // write canpos configuration
        if((CanposState == CANPOS_COUNT_WRITE_CONFIG_ACK))
        {
            scancode = collecting_canpos_write_config_ack(scancode);
        }
        // write flash
        else if((CanposState == CANPOS_COUNT_FLASH_ACK))
        {
            scancode = collecting_canpos_flash_ack(scancode);
        }
        // Collect Version Info
        else if((CanposState == CANPOS_COLLECT_VER_INFO_START))
        {
            scancode = collecting_canpos_ver_info(scancode);
        }
        //canpos ack
        else if((CanposState == CANPOS_COLLECT_NORMAL_ACK))
        {
            scancode = collecting_canpos_normal_ack(scancode);
        }
        // handle canpos new key 105 106
        else
        {
      	    scancode = check_canpos_key_105_106(scancode);
        }
    }


	/* *******************************************************************
	 * At this point we have a scancode that we haven't stuffed into a
	 * buffer.  This is either a valid scancode or an END OF something
	 * scancode.  If it is an END OF something scanCODE, we need to take
	 * some action and ground the scancode.  If it is a normal keystroke
	 * scanCODE, we will pass it on to the Boca code for handling.
	 * *******************************************************************
	 */
	switch (scancode) {
	case 0xFF:
		scanFlag = 1;
		break;

	case SCANCODE_DELETE:
	case SCANCODE_DELETE_UP:
		if (pPosKbdData->fTrap & PosTRAP_CTRL_ALT_DELETE) {
			if ((pPosKbdData->ShiftFlags & CTRL_DOWN) &&
			    (pPosKbdData->ShiftFlags & ALT_DOWN)) {
				/* *******************************************
				 * At least one ctrl and one alt key are down,
				 * so ground the keystroke.
				 * *******************************************
				 */
				scancode = 0;
				scanFlag = 1;
			}
		}
		break;

	case SCANCODE_ESC:
	case SCANCODE_ESC_UP:
		if ((pPosKbdData->ShiftFlags & CTRL_DOWN) &&
		    (pPosKbdData->fTrap & PosTRAP_CTRL_ESC)) {
			scancode = 0;
			scanFlag = 1;
		}

		if ((pPosKbdData->ShiftFlags & ALT_DOWN) &&
		    (pPosKbdData->fTrap & PosTRAP_ALT_ESC)) {
			scancode = 0;
			scanFlag = 1;
		}
		break;

	case SCANCODE_HOME:
	case SCANCODE_HOME_UP:
		if ((pPosKbdData->ShiftFlags & ALT_DOWN) &&
		    (pPosKbdData->fTrap & PosTRAP_ALT_HOME)) {
			scancode = 0;
			scanFlag = 1;
		}
		break;

	case SCANCODE_TAB:
	case SCANCODE_TAB_UP:
		if ((pPosKbdData->ShiftFlags & ALT_DOWN) &&
		    (!(pPosKbdData->ShiftFlags & SHIFT_DOWN)) &&
		    (pPosKbdData->fTrap & PosTRAP_ALT_TAB)) {
			scancode = 0;
			scanFlag = 1;
		}
		if ((pPosKbdData->ShiftFlags & ALT_DOWN) &&
		    (pPosKbdData->ShiftFlags & SHIFT_DOWN) &&
		    (pPosKbdData->fTrap & PosTRAP_ALT_SHIFT_TAB)) {
			scancode = 0;
			scanFlag = 1;
		}
		break;

	case SCANCODE_CAPS_LOCK:
		if (pPosKbdData->fTrap & PosTRAP_CAPS_LOCK) {
			scancode = 0;
			scanFlag = 1;
			pPosKbdData->ShiftFlags |= CAPS_LOCK_DOWN;
		}
		break;

	case SCANCODE_NUM_LOCK:
		if (pPosKbdData->fTrap & PosTRAP_NUM_LOCK) {
			scancode = 0;
			scanFlag = 1;
			pPosKbdData->ShiftFlags |= NUM_LOCK_DOWN;
		}
		break;

	case SCANCODE_SCROLL_LOCK:
		if (pPosKbdData->fTrap & PosTRAP_SCROLL_LOCK) {
			scancode = 0;
			scanFlag = 1;
			pPosKbdData->ShiftFlags |= SCROLL_LOCK_DOWN;
		}
		break;

	case SCANCODE_CAPS_LOCK_UP:
		if ((pPosKbdData->fTrap & PosTRAP_CAPS_LOCK) &&
		    (pPosKbdData->ShiftFlags & CAPS_LOCK_DOWN)) {
			scancode = 0;
			scanFlag = 1;
			pPosKbdData->ShiftFlags &= ~CAPS_LOCK_DOWN;
		}
		break;

	case SCANCODE_NUM_LOCK_UP:
		if ((pPosKbdData->fTrap & PosTRAP_NUM_LOCK) &&
		    (pPosKbdData->ShiftFlags & NUM_LOCK_DOWN)) {
			scancode = 0;
			scanFlag = 1;
			pPosKbdData->ShiftFlags &= ~NUM_LOCK_DOWN;
		}
		break;

	case SCANCODE_SCROLL_LOCK_UP:
		if ((pPosKbdData->fTrap & PosTRAP_SCROLL_LOCK) &&
		    (pPosKbdData->ShiftFlags & SCROLL_LOCK_DOWN)) {
			scancode = 0;
			scanFlag = 1;
			pPosKbdData->ShiftFlags &= ~SCROLL_LOCK_DOWN;
		}
		break;

	case SCANCODE_ROL:
		scancode = 0;
		scanFlag = 1;
		keyBufferIndex = 0;

		for (i = 0; i < sizeof(keyBuffer); ++i) {
			keyBuffer[i] = 0;
		}

		pPosKbdData->PosKbdStatus.Index = 0;     
		fStatus = 0;                             
		pPosKbdData->fStatus = 0;                
		pPosKbdData->MSRData.Index = 0;          

		/* wake up the process */
		cmd_busy = 0;
		wake_up_interruptible(&wq);

		/* notify aipctrl pid. */
		if (asyncEnabled) {
			kill_proc(asyncPID, SIGUSR2, 1);
		}
		break;

	case SCANCODE_ACK:
	case 0xAA:		       /* BAT complete */
	/* case 0xEE: */	       /* ECHO, key128 need to get remapped
					* to 0x59/0xd9
					*/
		scanFlag = 1;
		break;

	case SCANCODE_RESEND:
		scanFlag = 1;
		break;

	case SCANCODE_BEG_KBD_STATUS:
		pPosKbdData->PosKbdStatus.Index = 0;
		fStatus &= ~WAITING_FOR_KBD_STATUS;
		pPosKbdData->fStatus &= ~WAITING_FOR_KBD_STATUS;
		fStatus |= COLLECTING_KBD_STATUS;
		pPosKbdData->fStatus |= COLLECTING_KBD_STATUS;
		scancode = 0;
		scanFlag = 1;
		break;

	case SCANCODE_END_KBD_STATUS:
		/* ***********************************************************
		 * The same value is used for end of keyboard status and for
		 * the extended key break sequence of scan set 2.  We will
		 * assume that if we haven't seen a 0xF1 come by that this
		 * is a break sequence.
		 * ***********************************************************
		 */
		if (fStatus & COLLECTING_KBD_STATUS) {
			fStatus &= ~(COLLECTING_KBD_STATUS);
			pPosKbdData->fStatus &= ~COLLECTING_KBD_STATUS;

			/*  Keyboard Status */
#if 0
			DPRINTK("Kbd Status 0x%x 0x%x 0x%x 0x%x\n",
			        pPosKbdData->PosKbdStatus.KBDStatus[0],
			        pPosKbdData->PosKbdStatus.KBDStatus[1],
			        pPosKbdData->PosKbdStatus.KBDStatus[2],
			        pPosKbdData->PosKbdStatus.KBDStatus[3]);
#endif

			/* Capture and save Dev Info Data */
			if (pPosKbdData->PosKbdStatus.KBDStatus[1] &
			    POS_KBD_DEV_INFO_MSG_PRESENT) {
				/* Need to mask upper bit of DeviceId, Features,
				 * CommandSet and EC level except when a field
				 * is F3h or F4h -- in those cases, the entire
				 * upper nibble should be masked.
				 */

				for (devInfoInd = 4; devInfoInd <= 8; devInfoInd++) {
					if ((pPosKbdData->PosKbdStatus.KBDStatus[devInfoInd] == 0xF3) ||
					    (pPosKbdData->PosKbdStatus.KBDStatus[devInfoInd] == 0xF4)) {
						pPosKbdData->PosKbdStatus.KBDStatus[devInfoInd] &= 0x0F;
					} else {
						pPosKbdData->PosKbdStatus.KBDStatus[devInfoInd] &= 0x7F;
					}
				}

				/* save Device Info into a structure from Status
				 * info
				 */
				memcpy(&pPosKbdData->PosKbdDeviceInfo,
				       &pPosKbdData->PosKbdStatus.KBDStatus[4],
				       5);

				DPRINTK("Kbd DevInfo 0x%x.0x%x.0x%x.0x%x.0x%x\n",
				       pPosKbdData->PosKbdDeviceInfo.deviceType,
				       pPosKbdData->PosKbdDeviceInfo.deviceID,
				       pPosKbdData->PosKbdDeviceInfo.Features,
				       pPosKbdData->PosKbdDeviceInfo.CommandSet,
				       pPosKbdData->PosKbdDeviceInfo.ECLevel);
			}

			/* Check if the code update is completed. */
			if (pPosKbdData->PosKbdStatus.KBDStatus[1] &
			    CODE_UPDATE_STORED) {
				DPRINTK("CodeUpdate Complete\n");

				/* Set code update completion flag bit. */
				pPosKbdData->CodeUpdate.State |= CODE_UPDATE_STORED;
				if (pPosKbdData->CodeUpdate.State & CODE_UPDATE_LAST_BLOCK_SENT) {
					pPosKbdData->fCodeUpdtDone = 1;

					DPRINTK("CodeUpdate State=0x%x\n",
						pPosKbdData->fCodeUpdtDone);
				}
			}

			pPosKbdData->PosKbdStatus.Index = 0;
			scancode = 0;
			scanFlag = 1;

			/* wake up the process */
			cmd_busy = 0;
			wake_up_interruptible(&wq);

			/* notify aipctrl pid. */
			if (asyncEnabled) {
				kill_proc(asyncPID, SIGUSR2, 1);
			}
		} else {
			/* ***************************************************
			 * If we are not collecting MSR status, then the 0xF2
			 * needs to be handled below.  Do nothing here.
			 * ***************************************************
			 */
		}
		break;

	case SCANCODE_BEG_MSR_STATUS:
		pPosKbdData->MSRData.Index          = 0;
		pPosKbdData->MSRData.MSRDataLength  = 0;
		fStatus                            &= ~WAITING_FOR_MSR_STATUS;
		pPosKbdData->fStatus               &= ~WAITING_FOR_MSR_STATUS;
		fStatus                            |= COLLECTING_MSR_STATUS;
		pPosKbdData->fStatus               |= COLLECTING_MSR_STATUS;
		scancode                            = 0;
		scanFlag                            = 1;
		break;

	case SCANCODE_END_MSR_STATUS:
		if (fStatus & COLLECTING_MSR_STATUS) {
			/* ************* IMPORTANT NOTICE ********************
			 * The simple check for 0xF6 as the end of MSR data seems
			 * to be valid.  The ISO standards for Track 1, 2 & 3
			 * define characters to be 4 or 6 bit bytes.  The JIS
			 * standard for the JIS-II track allow for 8 bit bytes.
			 * The value 0xF6 falls in a range defined as "undefined"
			 * by the spec.  We will be safe as long as nobody tries
			 * to define 0xF6 as a meaningful value.
			 * ***************************************************
			 */
			if (ECBitMSR & pPosKbdData->MSRData.MSRDataBuffer[0]) {
				ECLevelMSR = pPosKbdData->MSRData.MSRDataBuffer[2];
			} else if (0x01 & pPosKbdData->MSRData.MSRDataBuffer[0]
				  && OB_KBD == KbdHWIDs) {
				/* *********************************************
				 * For Outerbanks keyboards, the first bit of
				 * the first status byte set tells us that this
				 * is a DEVICE INFO response.  We need to save
				 * the information so that we can be queried
				 * later.
				 * *********************************************
				 */
				ECLevelMSR = MSRdevinfo.ecLevel;
			}

			/* Capture and save Dev Info Data */
			if (pPosKbdData->MSRData.MSRDataBuffer[0] &
			    POS_MSR_DEV_INFO_MSG_PRESENT) {
				/* First copy data from Msr Data buffer into
				 * Msr Status buffer.
				 */
				memcpy(&pPosKbdData->PosMsrStatus.MSRStatus,
				       &pPosKbdData->MSRData.MSRDataBuffer, 7);

				/* Need to mask bit 7 since keyboard sets it */
				pPosKbdData->PosMsrStatus.MSRStatus[2] &= 0x3F;
				pPosKbdData->PosMsrStatus.MSRStatus[3] &= 0x0F;
				pPosKbdData->PosMsrStatus.MSRStatus[4] &= 0x0F;
				pPosKbdData->PosMsrStatus.MSRStatus[5] &= 0x0F;
				pPosKbdData->PosMsrStatus.MSRStatus[6] &= 0x0F;

				/* save Device Info into a structure from Status
				 * info
				 */
				memcpy(&pPosKbdData->PosMsrDeviceInfo,
				       &pPosKbdData->PosMsrStatus.MSRStatus[2],
				       5);

				/* Reset MSR Data Length */
				memset(&pPosKbdData->MSRData, 0,
				       sizeof(pPosKbdData->MSRData));

				pPosKbdData->MSRData.MSRDataLength  = 0;


				DPRINTK("Msr DevInfo 0x%x.0x%x.0x%x.0x%x.0x%x\n",
				       pPosKbdData->PosMsrDeviceInfo.deviceType,
				       pPosKbdData->PosMsrDeviceInfo.deviceID,
				       pPosKbdData->PosMsrDeviceInfo.Features,
				       pPosKbdData->PosMsrDeviceInfo.CommandSet,
				       pPosKbdData->PosMsrDeviceInfo.ECLevel);
			}


			/* ***************************************************
			 * Save the data length and reset the index to 0.  Also,
			 * reset the COLLECTING_MSR_STATUS flag.
			 * ***************************************************
			 */
			pPosKbdData->MSRData.Index = 0;

			fStatus &= ~COLLECTING_MSR_STATUS;
			pPosKbdData->fStatus &= ~COLLECTING_MSR_STATUS;
			scancode = 0;
			scanFlag = 1;

			/* wake up the process */
			cmd_busy = 0;
			wake_up_interruptible(&wq);

			/* notify aipctrl pid. */
			if (asyncEnabled) {
				kill_proc(asyncPID, SIGUSR2, 1);
			}

		}
		break;

	default:
		/* ***********************************************************
		 * Don't do anything here.  We will handle just past this
		 * switch statement.  It is possible to get a scancode that
		 * fits into a case statement above and is not grounded and
		 * hence needs to be handled (ie 0xF2 when we have not
		 * received a 0xF1 yet).
		 * ***********************************************************
		 */
		break;
	}

	/*  Crystal Support:  
	 *    - simply send out the scancode without any sort of translation
	 *      since the double keys not supported.
	 *    - the nanpos keyboard key positions do not make any sense except
	 *      get lost in translation
	 *    - return scancode right here w/o any processing.
	 */
	if ((pPosKbdData->PosKbdDeviceInfo.deviceID == 0x08) && scancode) {
		*pScancode = scancode;
		rc = 0;
		return rc;

	}

	/* *******************************************************************
	 * If we still have a non-grounded scancode at this point, we will
	 * need to determine if we have a full keystroke yet.  If we do, we
	 * will set the key value for doublekey and remap processing.
	 * *******************************************************************
	 */
	if (scancode && (0 == scanFlag)) {
		/* ***********************************************************
		 * At this point, the scancode hasn't been grounded, so we will
		 * put it in the buffer.
		 * ***********************************************************
		 */
		if (keyBufferIndex >= 3)
			keyBufferIndex = 0;

		keyBuffer[keyBufferIndex++] = scancode;

		DPRINTK("keyBufferIndex=0x%x, keyBuffer[0]=0x%x, keyBuffer[1]=0x%x, keyBuffer[2]=0x%x\n",
			keyBufferIndex, keyBuffer[0], keyBuffer[1],
			keyBuffer[2]);

		if (keyBufferIndex == 1) {
			if ((pPosKbdData->fStatus & DOUBLE_KEYS_109_108) &&
			    ((scancode == 0x1C) || (scancode == 0x9C)))
            {
                key = E0Table[KBD][scancode];
			}
            else
				key = ScanSet1Table[scancode];

			if (key) {
				fBreak = scancode > 0x7F ? (unsigned char) 1 : (unsigned char) 0;
			} else {
				scancode = 0;
			}

		} else if (0xE2 == keyBuffer[0]) {
			if (0xF2 != scancode) {
				if ((0xD9 <= scancode) && (scancode <= 0xEE)) {
					key = E2Table[scancode - 0xD9];
					fBreak = (0xF2 == keyBuffer[1] ? (unsigned char) 1: (unsigned char) 0);
				} else {
					scancode = 0;
				}
			}
		}

		if ((keyBuffer[0] == 0xE0) || (keyBuffer[0] == 0xE1)) {
			scancode = keyBuffer[0];
			keyBufferIndex = 0;
			key = 0;
		} else if ((scancode == 0xE2) || (scancode == 0xF2)) {
			/* don't reset keyBufferIndex, expecting more
			 * scancodes
			 */
            scancode = 0;
			key = 0;
		}
	}

	/* *******************************************************************
	 * If we still haven't grounded the keystroke, check for double key.
	 * *******************************************************************
	 */
	if (key) {
		/* ***********************************************************
		 * For now, make sure that we have explicitly set this value.
		 * Once the code is bug-free, we can choose an initial value and
		 * only change it if necessary.
		 * ***********************************************************
		 */
        key = DoubleKeyProcessing(key, fBreak);

		if (0 == key) {
			/* ***************************************************
			 * Clear the scancode buffer
			 * ***************************************************
			 */
            while (keyBufferIndex)
				keyBuffer[--keyBufferIndex] = 0;

			scancode = 0;
		}
	}

	/* *******************************************************************
	 * If we still have a key, clear out the scancode buffer and replace
	 * what's there with the desired scancodes for the keystroke.  This
	 * may very well have changed if we have replaced a secondary double
	 * key value with the primary.  In the future we may want to do this
	 * only if we know that the double key substitution has been done.
	 * *******************************************************************
	 */
	if (key) {
		/* ***********************************************************
		 * Clear the scancode buffer
		 * ***********************************************************
		 */

        while (keyBufferIndex)
			keyBuffer[--keyBufferIndex] = 0;

		/* ***********************************************************
		 * Copy in the scancodes.
		 * ***********************************************************
		 */
		while (ScancodesToGo[key][keyBufferIndex]) {
			keyBuffer[keyBufferIndex] = ScancodesToGo[key][keyBufferIndex];
#if 0
			DPRINTK("scTogo ==> keyBuf[keyBufferIndex] = 0x%x\n",
				keyBuffer[keyBufferIndex]);
#endif

			++keyBufferIndex;
		}

		/* ***********************************************************
		 * If this is a break sequence, we either need to set the high
		 * bit (scan set 1) or shift the second byte into the third
		 * position and add a 0xF2 as the new second byte.
		 * ***********************************************************
		 */
		if (fBreak) {
			keyBuffer[keyBufferIndex - 1] |= 0x80;
		}

		/* ***********************************************************
		 * First, if the key is a POS unique key and we are trapping off
		 * POS unique keys, then ground it here.  Second, we need to
		 * remap the POS unique key scancodes to the Mark McClanahan
		 * defined ones.
		 * ***********************************************************
		 */

		/* ***********************************************************
		 * If we are trapping off POS unique keys and this is one of
		 * them, ground the key.
		 * ***********************************************************
		 */
		if (pPosKbdData->fTrap & PosTRAP_POS_SPECIFIC &&
		    POSSpecific[key]) {
			key = 0;

            while (keyBufferIndex)
				keyBuffer[--keyBufferIndex] = 0;
		}

		/* ***********************************************************
		 * Allow the remap function to change the scancodes.  This is
		 * used to set the Intrepid defined scancodes for POS unique
		 * keys.  Later, it can be used to do general remapping of
		 * keys.
		 * ***********************************************************
		 */
		if (key) {
			/* ***************************************************
			 * Remap POS specific scancodes.  Remapping is needed
			 * to ensure the OS passes the scancodes to applicatoins.
			 * ***************************************************
			 */

            unsigned char remapKey = RemapTable[key];
            if (remapKey) {
				while (keyBufferIndex)
					keyBuffer[--keyBufferIndex] = 0;

				keyBuffer[0] = fBreak ? remapKey | (unsigned char) 0x80 : remapKey;
				keyBufferIndex = 1;
			}

		}
#if 0
		DPRINTK("5) key = 0x%x, buf[0] = 0x%x, Index1 = 0x%x\n",
			key, keyBuffer[0], keyBufferIndex);
#endif

		scancode = keyBuffer[0];
		keyBufferIndex = 0;
	}

	/* We need to maintain the shift state of SHIFT, CTRL, ALT keys.
	 * This is needed to trap certain keys.
	 */
	if (scancode == (unsigned char) 0xFF) {
		*scanState = OK;
	} else {
        switch (*scanState) {
		case OK:
            if (scancode == (unsigned char) 0xE0) {
				*scanState |= gotE0;
				break;
			} else if (scancode == (unsigned char) 0xE1) {
				*scanState |= gotE1;
				break;
			}

		case gotE0:
		case gotE1:
			/* Save Shift States for Pos Keyboard */
			SetShiftKeyState(scancode, scanState);
			/* Reset the state to OK. */
			*scanState = OK;
			break;
		}
	}

	/* setup rc and scancode to be pass. */
	if (scancode == 0) {
		/* scancode is consumed by us */
		*pScancode = 0;
		rc = 1;
	} else {
		/* return scancode for further processing */
		*pScancode = scancode;
		rc = 0;
	}

	/* handle new canpos key 105-106 */
	if (CanposKeyboard == 1) {
		if (canpos_key_105_106 != 0) {
			DPRINTK("canpos key_105-106");
			scancode = canpos_key_105_106;
			keyBuffer[0] = scancode;
			*pScancode = scancode;
			keyBufferIndex = 0;
			rc = 0;
			canpos_key_105_106 = 0;
		}
	}

	DPRINTK("rc=0x%x, scancode=0x%x, Index=0x%x, keyBuffer[0]=0x%x, keyBuffer[1]=0x%x, keyBuffer[2]=0x%x\n",
		rc, *pScancode, keyBufferIndex, keyBuffer[0], keyBuffer[1],
		keyBuffer[2]);

	return rc;
}


/* :H1.CanposPosKbdSendData() - Send data to keyborad (canpos).
 * ***********************************************************************
 *
 * Function Name:  CanposPosKbdSendData()
 *
 * Purpose:        To send data to keyboard (canpos).
 *
 * Description:    A routine to send data to canpos keyboard (larger buffer).
 *
 * Input:          unsigned char * buffer.
 *                 size of buffer
 *
 * Output:         returns number of bytes transferred.
 *                 0 - success
 *                 1 - failure.
 *
 * Notes:
 *
 * ***********************************************************************
 */

static int CanposPosKbdSendData(unsigned char *buf, unsigned int size)
{
	int rc = 0;
    int cmd_retries = 4;
    int xferred = 0, xfer_retry = size;
    

	while (cmd_busy && cmd_retries) {
		DPRINTK("going to sleep\n");

		set_current_state(TASK_INTERRUPTIBLE);
		schedule_timeout(5);
		--cmd_retries;

		DPRINTK("return from sleep\n");
	}

	/* Call pc_key driver's send routine to send data. */
	cmd_busy = 1;

    do {
		DPRINTK("sending %d bytes to kbd\n", size);

		/* size: number of retires and size of buffer */		
        rc = pos_filter.write_data(buf, size, &xferred, size,1000);

		if (rc != 0) {
			DPRINTK("pos_filter.write_data returned error %d\n", rc);

			/* the kernel patch misses ACK, and does
			 * not increment the xferred count
			 */
			if (rc == (-EINTR))
				xferred++;

			if (xferred < size) {
				buf += xferred;
				size -= xferred;
				xferred = 0;
			}
		}
	} while (xferred < size && xfer_retry--);

	/* Go to sleep.  The wait queue is awakened in pos_kbd_filter() routine
	 * when command is complete i.e. receipt of
	 *
	 *   -  0xF2 end of keyboard status OR
	 *   -  0xF6 end of msr status      OR
	 *   -  0xF7 ROL in response to keyboard reset.
	 */

    DPRINTK("Completed the call-waiting for timeout.\n");

    if (rc != 0)
		cmd_busy = 0;

	if (cmd_busy != 0)
        interruptible_sleep_on_timeout(&wq, 100);

	DPRINTK("exit, rc=0x%x %s\n", rc, rc == 0 ? "Success" : "Failed");

	return rc;

}


/* todo -- add prolog
 * collect canpos configuration data
 * set CanposState = CANPOS_COLLECT_READ_CONFIG_DATA upon enter this routine
 * when collecting done, exit with CanposState = CANPOS_READY
 * 
 */

static unsigned char colleting_canpos_read_config_data(unsigned char sc)
{
	unsigned char rc = sc;

	if (CanposState == CANPOS_COLLECT_READ_CONFIG_DATA) {
		if (rc == 0xFA)
			canposConfigAck++;

		/* save it all in buffer */
		canpos_config.Data[CanposConfigBytesCollected++] = sc;

		if (CanposConfigBytesCollected >= W_CONFIG_READ_RETURN_SIZE - W_CONFIG_READ_RETURN_OFFSET) {
			DPRINTK("Canpos collect config data: CanposState=CANPOS_READY\n");
			CanposState = CANPOS_READY;
		}

		if (rc != 0xfa)
			rc = 0;		/* pass only ack character back */
	}

	return rc;
}


/* TODO:  Function header
 * collecting normal ack
 * return 0 if scancode being filter out
 * return scancode otherwise
 */

unsigned char collecting_canpos_normal_ack(unsigned char sc)
{

    if (sc == 0x02) {
        DPRINTK("CANPOS_COLLECT_NORMAL_ACK: Make of Key Press 1.\n");
        sc =  0x00;
        scanFlag =1;
    }
    else if(sc == 0x82) {
        /* got keypress of 1 sequence */
        DPRINTK("CANPOS_COLLECT_NORMAL_ACK: key-of-1: sc=0x%x\n", sc);
        sc =  0x00;
        scanFlag =1;
        CanposState = CANPOS_GOT_CONFIG_KEY_1;
    }
    return sc;
}

/*TODO:  Function header
 * return 0 if scancode being filter out
 * return scancode otherwise
 * this routine trap the scancode of key-press 1
 */

static unsigned char collecting_canpos_write_config_ack(unsigned char sc)
{
    if (canposConfigAck<1281) {
        /* wait for acks */
        canposConfigAck++;
        //Ack for E5
        if (sc != 0xfa && canposConfigAck==1281 ) {
        IPRINTK("CANPOS_COLLECT_WRITE_CONFIG_ACK: ACk for E5.\n");
        sc =  0xFA;
    }
    } else if (sc == 0x82||sc ==0xaa) {
        /* got keypress of 1 sequence */
        IPRINTK("CANPOS_COLLECT_WRITE_CONFIG_ACK: key-of-1: sc=0x%x\n", sc);
        if(sc ==0x82)
            CanposState= CANPOS_GOT_CONFIG_KEY_1;
        sc =  0;
        scanFlag =1;
    }
    return sc;
}



/*TODO:  Function header
 * collecting ack during a canpos write flash update data
 * return 0 if scancode being filter out
 * return scancode otherwise
 */

static unsigned char collecting_canpos_flash_ack(unsigned char sc)
{
    if (canposFlashAck<(flashSize+3)) {
        /* wait for acks */
        canposFlashAck++;
        scanFlag =1;
        //Ack for the Last Flash Data Byte i.e. LRC
        if (sc == 0xFE && canposFlashAck==(flashSize+3) ) {
            DPRINTK("CANPOS_COLLECT_FLASH_ACK: ACk for LRC.\n");
            sc =  0xFA;
            scanFlag =1;
        }
    }
    else if (sc == 0x02) {
        /* got keypress of 1 sequence */
        DPRINTK("CANPOS_COLLECT_FLASH_ACK: Make code  for Key1: sc=0x%x\n", sc);
        sc =  0;
        scanFlag =1;
    }else if (sc == 0x82 || sc ==0xaa) {
        /* got keypress of 1 sequence */
        DPRINTK("CANPOS_COLLECT_FLASH_ACK: key-of-1: sc=0x%x\n", sc);
        if(sc==0x82)
            CanposState = CANPOS_GOT_CONFIG_KEY_1;
        sc =  0;
        scanFlag =1;
    }

    return sc;
}



/* wake up the process */
static void wake_up_canpos_process(void)
{
	/* wake up the process */
	cmd_busy = 0;
	wake_up_interruptible(&wq);

	/* notify aipctrl pid. */
	if (asyncEnabled) {
		kill_proc(asyncPID, SIGUSR2, 1);
	}
}


static unsigned char check_canpos_key_105_106(unsigned char scancode)
{
	canpos_key_105_106 = 0;

	if (scancode == 0x4a ||
	    scancode == 0xca ||
	    scancode == 0x4e ||
	    scancode == 0xce) {
		DPRINTK("canpos-key-105-106: 0x%x\n", scancode);

		canpos_key_105_106 = scancode;
	}

	return scancode;
}

#ifdef CANPOS
unsigned char collecting_canpos_ver_info(unsigned char sc)
{

    if(sc==0xFA)
        scanFlag =1;
    else
    {
        verinfo[verinfosize] = sc;
        verinfosize++;
        scanFlag = 1;
        sc =0;
        if(verinfosize==VER_INFO_SIZE)
        {
            DPRINTK("CANPOS_COLLECT_VER_INFO: Ver Info Collected Successfully.\n");
            verinfosize = 0;
            CanposState = CANPOS_COLLECT_VER_INFO_COLLECTED;
        }
    }

    return sc;
}
#endif

module_init(kb_init);
module_exit(kb_exit);


